from argparse import ArgumentParser
import yaml
import torch
import tifffile
from os.path import join, basename, splitext, isdir, normpath
from os import mkdir
from glob import glob
from numpy import savez
from omegaconf import OmegaConf

from nagini3D.models.model import DeepBioSnake
from nagini3D.data.th_optim_set import OptimSet


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("-i", "--input", help="str: path to the set dir")
    parser.add_argument("-m", "--model", help="str: path to the model config and weights directory")
    parser.add_argument("-w", "--weights", default="best.pkl", help="str: precise the name of the weights file (only name, not path) if you want a different one from best epoch")
    parser.add_argument("-t", "--tiles", default="1,1,1", help="(int,int,int): number of slices along each dimensions (ex : 1,1,1), if 'cuda out of memory' occurs")

    args = parser.parse_args()

    input_path = args.input
    model_dir = args.model 
    weigth_file = args.weights
    nb_tiles = [int(x) for x in args.tiles.split(",")]

    model_name = basename(normpath(model_dir))


    device = "cuda" if torch.cuda.is_available() else "cpu"

    with open(join(model_dir, "config.yaml"), 'r') as file:
        cfg = yaml.safe_load(file)

    M1 = cfg["settings"]["M1"]
    M2 = cfg["settings"]["M2"]
    r_mean = cfg["settings"]["r_mean"]

    weight_path = join(model_dir, weigth_file)

    model = DeepBioSnake(unet_cfg=cfg["model"], P = 101, M1 = M1, M2 = M2, save_path=".", device=device)
    model.load_weights(weight_path)

    optim_set = OptimSet(input_path)

    ths = model.optimize_thresholds(optim_set=optim_set, r_mean=r_mean, nb_tiles=nb_tiles, maxiter=10)

    OmegaConf.save(config=ths, f=join(model_dir, "thresholds.yaml"))


import torch
from torch import nn as nn
from torch.nn import functional as F
from functools import partial


class Level(nn.Module) :
    def __init__(self, ft_in, *, fst_c=None, down, scale_factor=2, convl, transit, convr, up, **kwargs) :
        super().__init__()
        self.down = down(in_ch = ft_in, out_ch = ft_in, scale_factor = scale_factor, **kwargs)
        self.convl = convl(ft_in if fst_c is None else fst_c, ft_in*2, **kwargs)
        self.transit = transit(ft_in*2, ft_in*2)
        self.convr = convr(ft_in*4, ft_in*2, **kwargs)
        self.up = up(ft_in*2, ft_in, scale_factor=scale_factor, **kwargs)

    def forward_down(self, x) :
        """Run the downard phase of the given level
        Parameters
        ----------
            x : input from previous level ( b, c, w, h)

        Returns
        -------
            x : output for next level ( b, c*2, w // sf, h // sf)
            skip : skip connection for same level ( b, c*2, w // sf, h /sf)
        """
        x = self.down(x)
        x = self.convl(x)
        skip = self.transit(x)
        return x, skip

    def forward_up(self, x, skip) :
        """Run the downard phase of the given level
        Parameters
        ----------
            x : input from next level ( b, c*2, w // sf, h // sf)
            skip : skip connection for same level ( b, c*2, w // sf, h // sf)

        Returns
        -------
            x : output for previous level ( b, c, w, h)
        """
        if skip is None :
            o = x
        else :
            x = self.shape_adapt(x, skip.shape)
            o =  torch.cat([skip, x], axis=1)
        o = self.convr(o)
        return self.up(o)


    @staticmethod
    def shape_adapt(x, shape_skip):
        """
        Adaptation of the shape of x if necessary ( due to non even feature map shape)
        """
        pads = []
        for i in range(2, len(x.shape))[::-1] : # Pad from right to left
            diff = shape_skip[i] - x.shape[i]
            pads.extend([diff//2, diff - diff//2])
        return F.pad(x, pads)

class UNetClean(nn.Module):
    """
    Paper: `U-Net: Convolutional Networks for Biomedical Image Segmentation
    <https://arxiv.org/abs/1505.04597>`_

    Paper authors: Olaf Ronneberger, Philipp Fischer, Thomas Brox

    Args:
        num_classes: Number of output classes required
        input_channels: Number of channels in input images (default 3)
        num_layers: Number of layers in each side of U-net (default 5)
        features_start: Number of features in first layer (default 64)
        kwargs :
            bilinear: Whether to use bilinear interpolation or transposed convolutions (default) for upsampling.
            train_bn : Whether to use accumulated batch parameters ( "trained" ) or per batch values
            inner_normalisation : Type of normalisation to use ['InstanceNorm', 'BatchNorm', 'None']
    """

    def __init__(
        self,
        num_classes: int,
        input_channels: int = 3,
        num_layers: int = 5,
        features_start: int = 64,
        **kwargs
    ):


        super().__init__()
        mLevel = partial(Level, **self.lalevels)
        self.num_layers = num_layers
        self.num_classes = num_classes
        self.features_start = features_start
        print(f'Num layers : {self.num_layers} Features Start : {features_start} Padding Mode : {kwargs["padding_mode"]}')
        assert features_start % 2 == 0, 'Feature start need to be even'
        l = {}
        l['level_0'] = mLevel(features_start // 2, fst_c=input_channels, down=nn.Identity, up=nn.Identity, **kwargs) # first level
        for k in range(1, num_layers-1):
            l[f'level_{k}'] = mLevel(features_start * 2**(k-1), **self.lalevels, **kwargs)
        l[f'level_{num_layers-1}'] = mLevel(features_start * 2**(num_layers-2), convr=nn.Identity,
                                      transit=self.lalevels.get('transit_bottleneck',self.lalevels['transit']), **kwargs) # last level

        l['final_layer'] = self.lalevels['final_layer'](features_start, num_classes, kernel_size=1) # Final Output

        self.layers = nn.ModuleDict(l)

    def forward(self, x):
        """
        Params :
            x : model input ( b, c, I, J)
        Returns:
            Segmentation : model segmentation ( b, L, I , J)
            Hidden Features : list with skip hidden representation ( num_layers, b, Ls, Is, Js)
        """
        bottleneck, skips = self.encode(x)
        x = self.decode(bottleneck, skips)
        out = self.layers['final_layer'](x)
        hidden_feats = skips + [bottleneck]

        return out, {'HiddenV' : hidden_feats}

    def encode(self, x) :
        """"
        Encode the input x and returns the bottleneck and skips.
        Each skip is the result of a "downsampling" and "transit" operation
        Params :
            x : model input ( b, c, I, J)
        Returns :
            bottleneck (b, Ls, Is, Js) : Bottom feature map
            skips : [b, Lk, Ik, Jk]*(num_layers-1) : skip connections
        """
        # Top to Bottom
        skips = []
        for k in range(self.num_layers-1) :
            x , skip = self.layers[f'level_{k}'].forward_down(x)
            skips.append(skip)
        _, bottleneck = self.layers[f'level_{self.num_layers-1}'].forward_down(x)
        return bottleneck, skips

    def decode(self, bottleneck, skips) :
        """
        Decode starting from hidden feats. Use the given skip connections
        Params :
            bottleneck (b, Ls, Is, Js) : Bottom feature map
            skips : [b, Lk, Ik, Jk]*(num_layers-1) : skip connections
        Returns :
            x (b, L_ft, I , J) : final feature map
        """
        x = self.layers[f'level_{self.num_layers-1}'].forward_up(bottleneck,None)
        for k in range(self.num_layers-1)[::-1] :
            x = self.layers[f'level_{k}'].forward_up(x, skips[k])
        return x

    def get_output_shape(self, in_dim) :
        """
        Return output shapes for this unet structure
        params : in_dim : input dimension ( c, w, h)
        """
        fk_in = torch.zeros((1, in_dim[0], in_dim[1], in_dim[2]))
        output, hidden_feats = self.forward(fk_in)
        return output.shape, hidden_feats.shape

    @staticmethod
    def add_specific_args(parser, prefix=''):
        #parser = ArgumentParser(parents=[parent_parser], add_help=False)
        prefix += 'unet.'
        parser.add_argument(f'--{prefix}train_bn', action='store_true')
        parser.add_argument(f'--{prefix}num_layers', type=int, default=3)
        parser.add_argument(f'--{prefix}features_start', type=int, default=32)
        parser.add_argument(f'--{prefix}padding_mode', type=str, choices=['zeros', 'reflect'], default='reflect')
        parser.add_argument(f'--{prefix}bilinear', action='store_true', help='bilinear upsampling in unet instead of convtranspose')
        parser.add_argument(f'--{prefix}inner_normalisation', '-inm', type=str, choices=['InstanceNorm', 'BatchNorm', 'None',
                                                                                     'cInstanceNorm2d', 'cBlockNorm3d',
                                                                                     'cBlockNorm3dSmooth',
                                                                                     'cBlockNorm3dSmoothV2'], default='InstanceNorm')

        return parser



import torch.nn as nn
import torch
import einops

class cInstanceNorm2d(nn.Module):

    def __init__(self, num_features, affine=False, track_running_stats=False, epsilon=1e-05) :
        """
        Applies Instance Normalisation on the 2 last dimensions of a N-d input.
        .. math::
            y = \frac{x - \mathrm{E}[x]}{ \sqrt{\mathrm{Var}[x] + \epsilon}} * \gamma + \beta
        """
        super().__init__()
        assert track_running_stats == False, f'track_running_stats = {track_running_stats} not implemented'
        assert affine == False, f'track_running_stats = {affine} not implemented'

        self.eps = epsilon
        self.num_features = num_features

    def forward(self, x) :
        """
        Applies layer normalisation.

        Args :
            x (N, C, *, H, W) : applies layer normalisation over (H, W).
                                Each channels (N, C, *) is independent
        Returns :
            normalised x (N, C, *, H, W)
        """
        assert x.shape[1] == self.num_features,\
         f'Expected (N, {self.num_features}, ...), got {x.shape}'

        r = (x - x.mean(axis=(-2, -1), keepdim=True))
        return r / torch.sqrt(x.var(axis=(-2, -1), keepdim=True, unbiased=False) + self.eps)

class cBlockNorm3d(nn.Module):

    def __init__(self, num_features, block_size=9, affine=False, track_running_stats=False, epsilon=1e-05) :
        """
        Normalise spatio temporal block with blocks of block_size dimension
        .. math::
            y = \frac{x - \mathrm{E}[x]}{ \sqrt{\mathrm{Var}[x] + \epsilon}} * \gamma + \beta
        """
        super().__init__()
        assert track_running_stats == False, f'track_running_stats = {track_running_stats} not implemented'
        assert affine == False, f'track_running_stats = {affine} not implemented'

        self.eps = epsilon
        self.num_features = num_features
        self.block_size = block_size

    def forward(self, x) :
        """
        Applies layer normalisation.

        Args :
            x (N, C, T, H, W) : applies layer normalisation over (block_size, H, W).
                                Each channel (N, C, *) is and each block is
        Returns :
            normalised x (N, C, T, H, W)
        """
        assert x.shape[1] == self.num_features,\
         f'Expected (N, {self.num_features}, ...), got {x.shape}'
        t = x.shape[2]
        if t % self.block_size != 0 :
            p = (self.block_size - t % self.block_size)
            x =nn.functional.pad(x, (0,0,0,0,p // 2 , p// 2+ p%2), 'reflect')
        num_groups = (x.shape[2] // self.block_size)
        x = einops.rearrange(x, 'b c t i j -> (b c) t i j', c=self.num_features)
        x = nn.functional.group_norm(x, num_groups)
        x = einops.rearrange(x, '(b c) t i j -> b c t i j', c=self.num_features)
        if t % self.block_size != 0 :
            x = x[:,:,p//2:-(p// 2+ p%2)]
        return x


class cBlockNorm3dSmooth(nn.Module):

    def __init__(self, num_features, block_size=9, affine=False, track_running_stats=False, epsilon=1e-05) :
        """
        Smooth normalisation on temporal dimension
        .. math::
            y = \frac{x - \mathrm{E}[x]}{ \sqrt{\mathrm{Var}[x] + \epsilon}} * \gamma + \beta
        """
        super().__init__()
        assert track_running_stats == False, f'track_running_stats = {track_running_stats} not implemented'

        self.affine = affine
        self.eps = epsilon
        self.num_features = num_features
        self.block_size = block_size
        self.conv_gaussian = nn.Conv1d(num_features,num_features, block_size,
                                       stride=1, padding='same',
                                       padding_mode='reflect',
                                       bias=False, groups=num_features)
        kernel = self.get_gaussian_kernel(block_size)
        self.conv_gaussian.weight.data = einops.repeat(kernel, 'ks -> c 1 ks', c=num_features).clone()
        self.conv_gaussian.weight.requires_grad_(False)
        if self.affine :
            self.weight =  torch.nn.Parameter(torch.ones(1, num_features, 1, 1, 1))
            self.bias =  torch.nn.Parameter(torch.zeros(1, num_features, 1, 1, 1))

    @staticmethod
    def get_gaussian_kernel(block_size, std=20) :
        #kernel = (-torch.arange(-(block_size+1)//2+1, (block_size+1)//2)**2 / (2*std)).exp()
        kernel = torch.ones(block_size)
        kernel /= kernel.sum()
        return kernel

    def forward(self, x) :
        """
        Applies layer normalisation.

        Args :
            x (N, C, T, H, W) : applies layer normalisation over (block_size, H, W).
                                Each channel (N, C, *) is and each block is
        Returns :
            normalised x (N, C, T, H, W)
        """
        assert x.shape[1] == self.num_features,\
         f'Expected (N, {self.num_features}, ...), got {x.shape}'
        # Apparently we don't need to detach the gradients
        # https://discuss.pytorch.org/t/batchnorm-and-back-propagation/65765/4
        # Step 1 : Remove running mean
        xmean = x.mean(axis=(-2,-1))
        x = (x - self.conv_gaussian(xmean)[...,None, None])

        # Step 2 : Remove running variance
        xvar = x.var(axis=(-2,-1))
        x_n = (x / torch.sqrt(self.conv_gaussian(xvar)[...,None, None]+self.eps))

        if self.affine :
            x = x * self.weight + self.bias
        return x


class cBlockNorm3dSmoothV2(nn.Module):

    def __init__(self, num_features, block_size=9, affine=False, track_running_stats=False, epsilon=1e-05) :
        """
        Smooth normalisation on temporal dimension
        .. math::
            y = \frac{x - \mathrm{E}[x]}{ \sqrt{\mathrm{Var}[x] + \epsilon}} * \gamma + \beta
        """
        super().__init__()
        assert track_running_stats == False, f'track_running_stats = {track_running_stats} not implemented'
        assert affine == False, f'affine = {affine} not implemented'


        self.affine = affine
        self.eps = epsilon
        self.num_features = num_features
        self.block_size = block_size
        if self.affine :
            self.weight =  torch.nn.Parameter(torch.ones(1, num_features, 1, 1, 1))
            self.bias =  torch.nn.Parameter(torch.zeros(1, num_features, 1, 1, 1))

    def forward(self, x) :
        """
        Applies layer normalisation.

        Args :
            x (N, C, T, H, W) : applies layer normalisation over (block_size, H, W).
                                Each channel (N, C, *) is and each block is
        Returns :
            normalised x (N, C, T, H, W)
        """
        assert x.shape[1] == self.num_features,\
         f'Expected (N, {self.num_features}, ...), got {x.shape}'
        # Apparently we don't need to detach the gradients
        # https://discuss.pytorch.org/t/batchnorm-and-back-propagation/65765/4
        # Step 1 : Remove running mean
        x_f = nn.functional.pad(x, (0,0,0,0, self.block_size  // 2 , self.block_size  // 2), 'reflect')
        x_f = x_f.unfold(2, self.block_size, 1)
        mu = einops.reduce(x_f, 'b c t i j k -> b c t 1 1', 'mean')
        std = (einops.reduce(x_f**2, 'b c t i j k -> b c t 1 1', 'mean') - mu**2 + 1e-8).sqrt()
        return (x - mu)/std



from argparse import ArgumentParser,ArgumentTypeError
import yaml
import torch
import tifffile
from os.path import join, basename, splitext, isdir, normpath
from os import mkdir
from glob import glob
from numpy import savez

from nagini3D.models.model import Nagini3D


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise ArgumentTypeError('Boolean value expected.')



if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("-i", "--input", help="str: path to the image directory")
    parser.add_argument("-m", "--model", help="str: path to the model config and weights directory")
    parser.add_argument("-w", "--weights", default="best.pkl", help="str: precise the name of the weights file (only name, not path) if you want a different one from best epoch")
    parser.add_argument("-o", "--output", help="str: path to the folder where to store inference results")
    parser.add_argument("-t", "--tiles", default="1,1,1", help="(int,int,int): number of splits along each dimensions (ex : 1,1,1), if the process exceed memory")
    parser.add_argument("-tt", "--thresholds", default="none", help="(float,float): precise the prob and nms thresholds to use if you don't want to appply the optimized ones (ex : 0.5,0.5)")
    parser.add_argument("-s", "--snake", type=str2bool, default=True, help="bool: weither to apply the snake refinement step")
    parser.add_argument("-ot", "--otsu", type=str2bool, default=True, help="bool: weither to apply otsu thresholding on image before snake refinement, more precise for sparse objects, less precise for dense objects.")
    parser.add_argument("-a", "--anisotropy", type=str, default="1,1,1", help="float,float,float: anisotrpoy ratio along each axis")

    args = parser.parse_args()

    input_path = args.input
    model_dir = args.model 
    weigth_file = args.weights
    result_path = args.output
    nb_tiles = [int(x) for x in args.tiles.split(",")]

    use_snake = args.snake
    use_otsu = args.otsu

    anisotropy = [float(x) for x in args.anisotropy.split(",")]

    model_name = basename(normpath(model_dir))

    device = "cuda" if torch.cuda.is_available() else "cpu"

    with open(join(model_dir, "config.yaml"), 'r') as file:
        cfg = yaml.safe_load(file)

    if args.thresholds == "none":
        with open(join(model_dir, "thresholds.yaml"), 'r') as file:
            thresholds = yaml.safe_load(file)
    else:
        thresholds = [float(x) for x in args.thresholds.split(",")]
        thresholds = {"prob" : thresholds[0], "nms" : thresholds[1]}

    M1 = cfg["settings"]["M1"]
    M2 = cfg["settings"]["M2"]
    r_mean = cfg["settings"]["r_mean"]

    weight_path = join(model_dir, weigth_file)

    model = Nagini3D(unet_cfg=cfg["model"], P = 101, M1 = M1, M2 = M2, save_path=result_path, device=device)
    model.load_weights(weight_path)

    imgs_path = glob(join(input_path,"*.tif"))

    proba_th = thresholds["prob"] 
    nms_th = thresholds["nms"]

    save_dir = join(result_path, model_name+f"_th_{round(proba_th,3)}_nms_{nms_th}")
    if not isdir(save_dir):
        mkdir(save_dir)

    mask_dir = join(save_dir, "mask")
    if not isdir(mask_dir): mkdir(mask_dir)
    

    for img_path in imgs_path:

        if device == "cuda":
            torch.cuda.empty_cache()

        img = tifffile.imread(img_path)
        img_name = basename(img_path)
        name = splitext(img_name)[0]

        print(f"Processing image : {img_name}")

        mask, proba, parameters, surfaces = model.inference(img, proba_th=proba_th, r_mean=r_mean, nb_tiles=nb_tiles,\
                                              nms_th=nms_th, optim_snake=use_snake, otsu=use_otsu, anisotropy=anisotropy)


        mask_path = join(result_path, img_name)
        proba_path = join(result_path, "proba_"+img_name)
        surfaces_path = join(result_path, "surf_"+name+".npz")
        tifffile.imwrite(mask_path, mask.cpu().numpy())
        tifffile.imwrite(proba_path, proba.cpu().numpy())
        savez(surfaces_path, points = surfaces["points"], facets = surfaces["facets"], values = surfaces["values"])


import numpy as np
from scipy.ndimage import binary_erosion, binary_dilation
from torch.nn.functional import conv3d
import torch
from random import randint
from scipy.signal.windows import gaussian


### Creation of the convolution kernel used for fast-marching computation

sampling_kernel = torch.zeros((19, 1, 3, 3, 3), dtype=torch.float32)
idx_no = [0,2,6,8,18,20,24,26]
count = 0
for p in range(27):
    if p not in idx_no:
        k, tmp = p%3, p//3
        j, i = tmp%3, tmp//3
        sampling_kernel[count, 0, i,j,k] = 1
        count += 1


def gkern(std = (3,3,3), th = 0.05):
    """Returns a 3D Gaussian kernel array."""
    std_v = np.array(std)
    std_x, std_y, std_z = std
    kerlen = (6*std_v).astype(int)
    kerlen += 1*(kerlen%2==0)
    lx, ly, lz = kerlen
    ker_x = gaussian(lx, std=std_x)[:, None, None]
    ker_y = gaussian(ly, std=std_y)[None, :, None]
    ker_z = gaussian(lz, std=std_z)[None, None, :]

    ker = ker_x*ker_y*ker_z

    return (ker>th)*ker


def farthest_point_sampling(contour_mask, nb_points = 51, device : str = "cpu", anisotropy = (1,1,1)):
    ax, ay, az = anisotropy
    crt_kern = torch.sqrt(torch.tensor([ax, 0, ax])[:,None,None]**2 + torch.tensor([ay, 0, ay])[None,:,None]**2 + torch.tensor([az, 0, az])[None,None,:]**2)
    idx_ker = torch.ones((27), dtype=bool)
    idx_ker[idx_no] = False
    add_ker = (crt_kern.flatten()[idx_ker]).to(device)[:,None,None,None]

    device_kernel = sampling_kernel.to(device)

    nx,ny,nz = contour_mask.shape
    i, j, k = 0, 0, 0

    # finding starting point
    while i<nx and contour_mask[i,j,k]==0:
        while j<ny and contour_mask[i,j,k]==0:
            while k<nz and contour_mask[i,j,k]==0:
                k+=1
            if k>=nz:
                j+=1
                k = 0
        if j>=ny:
            i+=1
            j=0

    sampling_list = [[i,j,k]]

    contour_tensor = torch.tensor(contour_mask, dtype=bool, device=device)

    # Adding points to the sampling list
    while len(sampling_list) < nb_points:

        old_fast_marching = torch.zeros_like(contour_tensor, dtype=torch.float32)
        new_fast_marching = old_fast_marching.clone()
        for i,j,k in sampling_list:
            new_fast_marching[i,j,k] = 1 # pseudo_inf

        # Computing fast marching algorithm
        while ((new_fast_marching - old_fast_marching)*contour_tensor).sum()!=0:
            old_fast_marching = new_fast_marching.clone()
            conv_step = (conv3d(new_fast_marching[None,...], device_kernel, padding="same"))
            new_fast_marching = ((conv_step + torch.where(conv_step > 0, add_ker, torch.inf)).min(dim=0)[0])
            new_fast_marching = torch.where((new_fast_marching<torch.inf)*contour_tensor, new_fast_marching, 0)
            #new_fast_marching = new_fast_marching.where(contour_tensor, 0)
            # new_fast_marching = torch.maximum(old_fast_marching, new_fast_marching)
        

        # Choosing one of the farthest point to the sampling subset
        maxi_list = (new_fast_marching==new_fast_marching.max()).nonzero() # prendre un indice random plutot que le premier
        i,j,k = maxi_list[randint(0,len(maxi_list)-1)]

        # Adding this point to the set
        sampling_list.append([i.item(),j.item(),k.item()]) 

    return sampling_list



distance_kernel = torch.zeros((6, 1, 3, 3, 3), dtype=torch.float32)
for idx, value in enumerate([4,10,12,14,16,22]):
    i, tmp = value%3, value//3
    j, k = tmp%3, tmp//3
    distance_kernel[idx, 0, i,j,k] = 1


def distance_to_center(mask):

    eroded = mask
    proba = 0
    
    while eroded.sum()>0:
        proba += eroded
        eroded = binary_erosion(eroded)

    return proba/proba.max()


def mask_to_contour(mask, mode  : str = "erosion"):
    if mode == "erosion":
        eroded = binary_erosion(mask)
        return mask - eroded
    if mode == "dilation":
        dilat = binary_dilation(mask)
        return dilat - mask

def compute_barycenter(mask, mesh):
    mx, my, mz = mesh
    N = mask.sum()

    # cell barycenter coordinates
    x = round((mask*mx).sum()/N) 
    y = round((mask*my).sum()/N)
    z = round((mask*mz).sum()/N)

    return (x,y,z)

def compute_radius(contour, mesh, barycenter):
    mx, my, mz = mesh
    N = contour.sum()
    x,y,z = barycenter

    # radius of the cell
    r = np.sqrt(((mx-x)**2+(my-y)**2+(mz-z)**2)*contour).sum()/N

    return r

def compute_sigma(mask, mesh, barycenter):

    mx, my, mz = mesh
    x, y, z = barycenter
    N = mask.sum()

    rx = (np.abs(mx-x)*mask).sum()/N 
    ry = (np.abs(my-y)*mask).sum()/N
    rz = (np.abs(mz-z)*mask).sum()/N
    
    return rx, ry, rz

def bound_box(mask, mesh):
    mx, my, mz = mesh
    
    x_unique = np.unique(mask*(mx+1))
    x_min, x_max = x_unique[1]-1, x_unique[-1]-1

    y_unique = np.unique(mask*(my+1))
    y_min, y_max = y_unique[1]-1, y_unique[-1]-1

    z_unique = np.unique(mask*(mz+1))
    z_min, z_max = z_unique[1]-1, z_unique[-1]-1

    return x_min, x_max, y_min, y_max, z_min, z_max


    


import torch
from random import shuffle

class PointCloudDisplay():

    def __init__(self, M1 , M2) -> None:
        self.M1 = M1
        self.M2 = M2


    def display_val(self, nb_cells, GT, pred, cp, K = 10): # K is the number of points displayed along the tangents
        
        tot_cells, nb_points, _ = GT.shape
        idx = [i for i in range(tot_cells)]
        shuffle(idx)
        idx = idx[:nb_cells]

        clouds_list = list()
        M1, M2 = self.M1, self.M2

        for i in idx:
            crt_GT = torch.concatenate((GT[i], torch.ones((nb_points, 1))),dim=-1)
            crt_pred = torch.concatenate((pred[i], 2*torch.ones((nb_points, 1))),dim=-1)
            reg_cp = torch.concatenate((cp[i,:M1*(M2-1)], 4*torch.ones((M1*(M2-1), 1))),dim=-1)

            poles = cp[i,M1*(M2-1):M1*(M2-1)+2]
            
            T_N = (poles[0][None,None,:] + cp[i,M1*(M2-1)+2:M1*(M2-1)+4].unsqueeze(1)*torch.linspace(0,1,K)[None,:,None]).reshape(-1,3)
            T_S = (poles[1][None,None,:] + cp[i,M1*(M2-1)+4:M1*(M2-1)+6].unsqueeze(1)*torch.linspace(0,1,K)[None,:,None]).reshape(-1,3)

            grad = torch.concatenate((torch.concatenate((T_N,T_S), dim=0),9*torch.ones(4*K,1)),dim=-1)

            clouds_list.append(torch.concatenate((crt_GT,crt_pred, reg_cp, grad), dim=0).numpy())

        return clouds_list
            



from argparse import ArgumentParser
import numpy as np
from tifffile import imread
import napari 


if __name__ == "__main__":

    parser = ArgumentParser()
    parser.add_argument("-p", "--points")
    parser.add_argument("-i", "--input")

    args = parser.parse_args()

    points_file = args.points
    img_file = args.input

    img = imread(img_file)[..., :200]

    contours = np.load(points_file)


    points = contours["points"]
    facets = contours["facets"]
    values = contours["values"]

    viewer = napari.view_labels(img, ndisplay=3)

    for point_cloud in points[0:50]:
        viewer.add_surface((point_cloud, facets, values), colormap="I Orange")

    napari.run()


import torch
from math import pi, cos

def create_interval_mask(t, lower_bound, upper_bound):
    """
        Return a binary with 1 where lower_bound <= t < upper_bound and 0 elsewhere.
    """
    return (lower_bound <= t)*(t<upper_bound)*1.


def create_exponential_spline(M):

    def basis_function_t(t):
        abs_t = torch.abs(t)

        mask_0 = create_interval_mask(abs_t, 0, 0.5)
        values_0 = mask_0*(torch.cos(2*pi*abs_t/M)*cos(pi/M) - cos(2*pi/M))

        mask_1 = create_interval_mask(abs_t, 0.5, 1.5)
        values_1 = mask_1*(torch.sin(pi*(3/2-abs_t)/M)*torch.sin(pi*(3/2-abs_t)/M))
        
        phi_M_t = 1/(1-cos(2*pi/M))*(values_0 + values_1)

        return phi_M_t

    return basis_function_t


def create_periodic_exponential_spline(M):

    basis_fct_aux = create_exponential_spline(M)

    # sufficient to use this support because for other indices the function is zero everywhere
    n_min = -1
    n_max = 1


    def basis_periodic_function(t):
        phi_t = 0
        for n in range(n_min,n_max+1):
            phi_t += basis_fct_aux(t-M*n)
        return phi_t

    return basis_periodic_function



def create_polynomials():

    def basis_function(t):
        t_2 = t*t
        mask_0 = create_interval_mask(t, -1, 0)
        values_0 = mask_0*(2-t_2*(3*t+5))

        mask_1 = create_interval_mask(t, 1, 2)
        values_1 = mask_1*(-(t-2)*(t-2)*(t-1))

        mask_2 = create_interval_mask(t, -2, -1)
        values_2 = mask_2*((t+1)*(t+2)*(t+2))

        mask_3 = create_interval_mask(t, 0, 1)
        values_3 = mask_3*((3*t-5)*t_2 + 2)

        phi_t = 0.5*(values_0 + values_1 + values_2 + values_3)

        return phi_t
    
    return basis_function




from torch.utils.data import Dataset
from os.path import join, basename, splitext
from glob import glob
import tifffile
import numpy as np
from random import randint
import torch
from typing import Any
from csbdeep.utils import normalize

from .data_aug_tools import generate_rotation_matrix, rotate_image

FLOAT_TYPE = np.float32
INT_TYPE = np.uint16

class TrainingSet(Dataset):

    def __init__(self, dataset_dir : str, patch_size : list, nb_points : int = 101, anisotropy_ratio : list = [1,1,1],
                 r_mean = None, data_aug : bool = True, cell_ratio_th = 0.0, **kwargs) -> None:
        super().__init__()

        self.images_dir = join(dataset_dir, "images")
        self.samplings_dir = join(dataset_dir, "samplings")
        self.masks_dir = join(dataset_dir, "masks")

        self.patch_size = patch_size
        self.nb_points = nb_points
        self.data_aug = data_aug

        self.is_isotropic = (anisotropy_ratio == None)
        self.anisotropy_ratio = np.array([[[1,1,1]]]) if self.is_isotropic else np.array(anisotropy_ratio)[None,None,:]

        self.imgs_list = sorted(glob(join(self.images_dir, "*.tif")))

        self.samplings_list = glob(join(self.samplings_dir, "*.npz"))

        self.cell_ratio_th = cell_ratio_th

        if r_mean != None:
            self.r_mean = r_mean

        else:
            r_mean = 0
            nb_cells = 0
            for sampling_file in self.samplings_list:
                radius = np.load(sampling_file)["radius"]
                n = len(radius)
                r_mean = nb_cells/(n+nb_cells)*r_mean + n/(n+nb_cells)*radius.mean()
                nb_cells += n

            self.nb_cells = nb_cells
            self.r_mean = float(r_mean)



    def __len__(self):
        return len(self.imgs_list)
    

    def crop_image(self, img) -> tuple:
        nx,ny,nz = img.shape 
        Mx, My, Mz = self.patch_size
        x_max, y_max, z_max = nx-Mx, ny-My, nz-Mz
        x, y, z = randint(0, x_max), randint(0, y_max), randint(0, z_max)
        return img[x:x+Mx, y:y+My, z:z+Mz], (x,y,z)

    def __getitem__(self, index) -> Any:
        img_path = self.imgs_list[index]
        img_name = basename(img_path)
        img_name_no_ext = splitext(img_name)[0]


        full_img = tifffile.imread(img_path)
        full_img = normalize(full_img, pmin=1, pmax=99.8, axis=(0,1,2), clip=True)

        full_proba = tifffile.imread(join(self.samplings_dir, img_name))
        full_mask = tifffile.imread(join(self.masks_dir, img_name))

        samplings_arrays = np.load(join(self.samplings_dir, img_name_no_ext+".npz"))

        full_samplings = samplings_arrays["samplings"]
        full_centers = samplings_arrays["centers"]

        Mx, My, Mz = self.patch_size

        mask, (x,y,z) = self.crop_image(full_mask)
        mask = mask.astype(INT_TYPE)

        if self.cell_ratio_th != 0: # avoid to compute the sum when one chooses the ratio equal to 0 (i.e. no verification)
            cell_ratio_in_mask = (mask>0).sum()/(Mx*My*Mz)
            while cell_ratio_in_mask < self.cell_ratio_th:
                mask, (x,y,z) = self.crop_image(full_mask)
                mask = mask.astype(INT_TYPE)
                cell_ratio_in_mask = (mask>0).sum()/(Mx*My*Mz)


        proba = full_proba[x:x+Mx, y:y+My, z:z+Mz].astype(FLOAT_TYPE)
        img = full_img[x:x+Mx, y:y+My, z:z+Mz].astype(FLOAT_TYPE)

        unique_lbls = np.unique(full_mask)[1:]
        lbls = np.arange(unique_lbls[-1]+1, dtype=int)
        lbls[unique_lbls] = np.arange(len(unique_lbls), dtype=int)


        if self.data_aug and self.is_isotropic:
            theta1, theta2, theta3 = randint(0,3), randint(0,3), randint(0,3)
            angles = (theta1, theta2, theta3)

            img = rotate_image(img, angles).copy()
            proba = rotate_image(proba, angles).copy()
            mask = rotate_image(mask, angles).copy()
        

        cell_voxels = mask.nonzero()
        voxels_coordinates = np.stack(cell_voxels).T
        voxels_idx = lbls[mask[cell_voxels]]
        voxels_proba = proba[cell_voxels]
        voxels_barycenters = (full_centers[voxels_idx] - np.array([x,y,z]))

        assert self.nb_points <= full_samplings.shape[1]
        tmp_samplings = full_samplings[voxels_idx,:self.nb_points] 

        
        if self.data_aug and self.is_isotropic:
            R = generate_rotation_matrix(theta1*np.pi/2, theta2*np.pi/2, theta3*np.pi/2)
            centering_shift = (self.patch_size[0]-1)/2
            voxels_barycenters = (R@(voxels_barycenters-centering_shift).T).T + centering_shift
            tmp_samplings = np.reshape(((R@np.reshape(tmp_samplings,(-1,3)).T).T), tmp_samplings.shape)


        voxels_samplings = (((tmp_samplings + voxels_barycenters[:,None,:]\
                             - voxels_coordinates[:,None,:]+0.5)*self.anisotropy_ratio-0.5)/self.r_mean).astype(FLOAT_TYPE)
    


        return {"image" : torch.tensor(img).unsqueeze(0), "proba" : torch.tensor(proba),
                "voxels_samplings" : torch.tensor(voxels_samplings), "voxels_proba" : torch.tensor(voxels_proba),
                "cell_voxels": cell_voxels}
    


def custom_collate(batch):
    return {
        "images" : torch.stack([item["image"] for item in batch]),
        "proba" : torch.stack([item["proba"] for item in batch]),
        "voxels_samplings" : torch.cat([item["voxels_samplings"] for item in batch]),
        "voxels_proba" : torch.cat([item["voxels_proba"] for item in batch]),
        "cell_voxels" : [item["cell_voxels"] for item in batch]
    }


# if __name__ == "__main__":
#     path = ""
#     anisotropy = [1,1,2]
#     dataset = TrainingSet(dataset_dir=path, patch_size=50, nb_points=61, anisotropy_ratio=anisotropy)

#     item = dataset[1]

#     img = item["image"]
#     proba = item["proba"]
#     samplings = item["voxels_samplings"]
#     vox_proba = item["voxels_proba"]
#     cell_vox = item["cell_voxels"]

#     gamma = (np.array(img.shape)[1:]/np.array(proba.shape))
#     coor = (np.stack(cell_vox).T)*gamma

#     N = samplings.shape[0]

#     points_to_plot = list()

#     random_idx = [k for k in range(N)]
#     from random import shuffle
#     shuffle(random_idx)

    


#     for i in random_idx[:5]:
#         s = samplings[i]
#         c = coor[i]
#         s_shifted = s*dataset.r_mean + c[None,:]*np.array([anisotropy])
        
#         # print(s)
#         points_to_plot.append(s_shifted)


#     # print(len(cell_vox[0])/(proba==0).sum())

    
    
#     import napari
#     from skimage.transform import rescale
#     img = rescale(img[0].numpy(), anisotropy)
#     viewer = napari.view_image(img, ndisplay=2)
#     #viewer.add_image(proba.numpy())

#     #viewer.add_points(coor, face_color="red", size=1)
#     viewer.add_points(np.concatenate(points_to_plot), face_color="red", size=1)
#     #viewer.add_points(coor*np.array([anisotropy]),size = 1, face_color="green")
    

#     napari.run()


import torch
import torch.nn as nn  

class SlicedWassersteinLoss(nn.Module):
  def __init__(self, nb_angles = 100, W_order=1, p_order =1, device = "cpu") -> None:
    super().__init__()
    self.nb_angles = nb_angles
    self.d = W_order
    self.device = device
    self.p_order = p_order

  def forward(self, input, target, proba):
    """
    - input: (batch, nb_points, 3)
    - target: (batch, nb_points, 3) 
    """

    angles = (torch.rand(1, 1, self.nb_angles, 3)*2-1).to(self.device)
    angles = angles/torch.linalg.vector_norm(angles, dim=-1, keepdim=True)

    inner_prod_input = (input.unsqueeze(-2)*angles).sum(dim=-1)
    inner_prod_target = (target.unsqueeze(-2)*angles).sum(dim=-1)

    sorted_input = torch.sort(inner_prod_input, dim=-2)[0]
    sorted_target = torch.sort(inner_prod_target, dim=-2)[0]

    diff = (sorted_input - sorted_target).abs()
    diff_pow = diff.pow(self.d)
    wasser = diff_pow.mean(dim=-2).pow(1/self.d).mean(dim=1)
    if self.p_order == 1:
      return (wasser*proba).sum()/(proba.sum())

    prob = proba.pow(self.p_order) 
    return (wasser*prob).sum()/(prob.sum())


import numpy as np
import torch
from math import pi, cos, sin

from .snake_tools import (create_exponential_spline, create_periodic_exponential_spline)

class SnakeSmoothSampler():
    def __init__(self, P, M1, M2=None, device = "cpu"):
        # Creation of Fibonacci Lattice 
        self.P = P
        assert self.P%2 == 1

        self.device = device

        golden_ratio = (np.sqrt(5)-1)/2

        N = P // 2
        i = np.arange(-N, N+1)
        lat = np.arcsin(2*i/P)/torch.pi + 0.5
        long = np.mod(i*golden_ratio, 1)

        self.u = torch.unsqueeze(torch.tensor(long),1) 
        self.v = torch.unsqueeze(torch.tensor(lat), 1)

        # Computation of sampling weights
        self.M1 = M1
        if M2 == None:
            self.M2=M1
        else:
            self.M2=M2

        i = torch.arange(0, self.M1).unsqueeze(0)
        j = torch.arange(-1, self.M2+2).unsqueeze(0)

        # Careful phi_1 = M1*phi but phi_2 = 2*M2*phi

        self.phi_per = create_periodic_exponential_spline(M1)
        self.phi_per_weights = self.phi_per(M1*self.u-i).unsqueeze(2)

        self.phi = create_exponential_spline(2*M2)
        self.phi_weights = self.phi(M2*self.v-j).unsqueeze(1)

        self.weights = (self.phi_per_weights*self.phi_weights).unsqueeze(0)[...,None].to(device)

        gamma = 2*(1-cos(2*pi/M1))/(cos(pi/M1)-cos(3*pi/M1))
        self.cM1 = gamma*torch.cos(2*torch.pi*i/M1)[...,None]
        self.sM1 = gamma*torch.sin(2*torch.pi*i/M1)[...,None]

        self.phi_0 = (cos(pi/(2*M2))-cos(pi/M2))/(1-cos(pi/M2))  #compute_phi_0(M2)
        self.phi_1 = (1-cos((pi)/(2*M2)))/(2*(1-cos(pi/M2))) # compute_phi_1(M2)
        self.phi_prime_1 = (-pi/(2*M2))*sin(pi/(2*M2))/(1-cos(pi/M2)) # compute_phi_prime_1(M2)

        self.fi_weights = self.init_fi_weights(self.phi_per_weights, self.phi_weights)[None,...,None].to(device=self.device)

        #print(f"Valeurs des vecteurs phi, 0 : {self.phi_0}, 1 : {self.phi_1}, 1' : {self.phi_prime_1}")

    def init_fi_weights(self, phi_1, phi_2):
        gamma = self.phi_1/self.phi_0
        Gamma = phi_2[...,2:-2]
        Gamma[...,0] += phi_2[...,0] - 2*gamma*phi_2[...,1]
        Gamma[...,-1] += phi_2[...,-1] - 2*gamma*phi_2[...,-2]

        fi_weights = phi_1*Gamma
        a,b,c = fi_weights.shape
        fi_weights = fi_weights.reshape(a,b*c)
        
        Gamma_N = phi_2[...,1]/self.phi_0
        Gamma_S = phi_2[...,-2]/self.phi_0

        Gamma_1N = phi_2[...,0] - gamma*phi_2[...,1]
        Gamma_1S = gamma*phi_2[...,-2] - phi_2[...,-1]

        alpha_k = self.cM1/(self.M2*self.phi_prime_1)
        beta_k = self.sM1/(self.M2*self.phi_prime_1)

        A_u = (alpha_k*phi_1).sum(dim=(-2))
        B_u = (beta_k*phi_1).sum(dim=(-2))
        
        return torch.cat((fi_weights, Gamma_N, Gamma_S, Gamma_1N*A_u,
                                     Gamma_1N*B_u, Gamma_1S*A_u, Gamma_1S*B_u), dim=-1)
    

    def up_sampling_uv(self, M1, M2):
        i = torch.arange(0, M1*(M2-1))
        k = i//(M2-1)
        l = i%(M2-1) + 1
        u = k/M1
        v = l/M2
        u = torch.cat((u, torch.tensor([0, 0.33, 0.66, 0, 0.33, 0.66])))
        eps = 1/(2*M2)
        v = torch.cat((v, torch.tensor([eps, eps, eps, 1-eps, 1-eps, 1-eps])))
    
    def up_sampling_uv(self, M1, M2):
        i = torch.arange(0, M1*(M2-1))
        k = i//(M2-1)
        l = i%(M2-1) + 1
        u = k/M1
        v = l/M2
        u = torch.cat((u, torch.tensor([0, 0.33, 0.66, 0, 0.33, 0.66])))
        eps = 1/(2*M2)
        v = torch.cat((v, torch.tensor([eps, eps, eps, 1-eps, 1-eps, 1-eps])))

        return u[...,None], v[...,None]

    def up_sample_transfer_matrix(self):
        u, v = self.up_sampling_uv(self.M1, self.M2)

        i = torch.arange(0, self.M1).unsqueeze(0)
        j = torch.arange(-1, self.M2+2).unsqueeze(0)

        phi_1 = self.phi_per(self.M1*u-i).unsqueeze(2)
        phi_2 = self.phi(self.M2*v-j).unsqueeze(1)

        return torch.linalg.inv(self.init_fi_weights(phi_1, phi_2)).to(self.device)
    
    def up_sample_sampling(self, fi, new_M1, new_M2):
        u, v = self.up_sampling_uv(new_M1, new_M2)

        i = torch.arange(0, self.M1).unsqueeze(0)
        j = torch.arange(-1, self.M2+2).unsqueeze(0)

        phi_1 = self.phi_per(self.M1*u-i).unsqueeze(2)
        phi_2 = self.phi(self.M2*v-j).unsqueeze(1)

        weights = self.init_fi_weights(phi_1, phi_2)[None,...,None].to(self.device)

        return (fi[:,None,...]*weights).sum(dim=-2)

    def free_parameters_to_cp(self, free_parameters):

        """
        - control_points: (batch, M1.(M2-1)+6, 3) #dans l'article ils disent +4 mais c'est très bizarre
            The control_points vector is composed of:
                - (M1x(M2-1),3) regular control points
                - (1, 3)  north pole control point
                - (1, 3)  south pole control point
                - (2, 3)  north pole derivative vector
                - (2, 3)  south pole derivative vector
        """
        cM1 = self.cM1.to(self.device)
        sM1 = self.sM1.to(self.device)

        M1, M2 = self.M1, self.M2
        batch_size = free_parameters.shape[0]

        reg_cp = free_parameters[:,:M1*(M2-1)].reshape(batch_size, M1, M2-1, 3)
        cp_N = free_parameters[:,M1*(M2-1)].unsqueeze(-2)
        cp_S = free_parameters[:,M1*(M2-1)+1].unsqueeze(-2)
        T_N1 = free_parameters[:,M1*(M2-1)+2].unsqueeze(-2)
        T_N2 = free_parameters[:,M1*(M2-1)+3].unsqueeze(-2)
        T_S1 = free_parameters[:,M1*(M2-1)+4].unsqueeze(-2)
        T_S2 = free_parameters[:,M1*(M2-1)+5].unsqueeze(-2)

        cp_i_1 = reg_cp[:,:,0]
        cp_i_M2_1 = reg_cp[:,:,-1]

        a = cp_i_1 + (T_N1*cM1 + T_N2*sM1)/(M2*self.phi_prime_1) 
        b = cp_N/self.phi_0  - self.phi_1*(a + cp_i_1)/self.phi_0 # careful, it is a "-"" not a "+" as said in the paper, there is a mistake in the proof

        d = cp_i_M2_1 - (T_S1*cM1 + T_S2*sM1)/(M2*self.phi_prime_1)
        c = cp_S/self.phi_0 - self.phi_1*(cp_i_M2_1 + d)/self.phi_0

        cp = torch.cat((a.unsqueeze(-2),
                        b.unsqueeze(-2),
                        reg_cp,
                        c.unsqueeze(-2),
                        d.unsqueeze(-2)), dim=-2)
        
        return cp.unsqueeze(1)
    

    def get_dckl_dfi(self):
        M1, M2 = self.M1, self.M2
        I = M1*(M2-1) + 6
        mid = torch.zeros((M1*(M2-1)+6, M1, M2-1))
        i = torch.arange(0,I-6)
        k = i//(M2-1)
        l = i%(M2-1)
        mid[i,k,l] = 1

        left = torch.zeros((M1*(M2-1)+6, M1, 2))
        right = torch.zeros((M1*(M2-1)+6, M1, 2))

        alpha = self.phi_1/self.phi_0
        R = (I - 6) - 1 # index where the fi that are not ckl starts 

        cM1 = self.cM1.squeeze()
        sM1 = self.sM1.squeeze()

        k = torch.arange(0, M1)
        left[(M2-1)*k,k,0] = 1
        left[R+3,:,0] = cM1/(M2*self.phi_prime_1)
        left[R+4,:,0] = sM1/(M2*self.phi_prime_1)

        left[(M2-1)*k,k,1] = -2*alpha
        left[R+1,:,0] = 1/self.phi_0
        left[R+3,:,1] = -alpha*cM1/(M2*self.phi_prime_1)
        left[R+4,:,1] = -alpha*sM1/(M2*self.phi_prime_1)

        right[(M2-1)*(k+1)-1,k,1] = 1
        right[R+5,:,1] = -cM1/(M2*self.phi_prime_1)
        right[R+6,:,1] = -sM1/(M2*self.phi_prime_1)

        right[(M2-1)*(k+1)-1,k,1] = -2*alpha
        right[R+2,:,0] = 1/self.phi_0
        right[R+5,:,1] = alpha*cM1/(M2*self.phi_prime_1)
        right[R+6,:,1] = alpha*sM1/(M2*self.phi_prime_1)

        return torch.cat((left,mid,right), dim = -1).to(self.device)

    def sample_snakes(self, fi):
        return (fi[:,None,...]*self.fi_weights).sum(dim=-2)
    

    def create_facets(self, n_u, n_v):
        i, j = torch.arange(n_u)[...,None], torch.arange(1,n_v-1)[None, ...]

        def f(u, v, width : int):
            return u*width+v
        
        s1 = f(i,j,n_v).reshape(n_u*(n_v-2))
        s2 = f(torch.remainder(i+1, n_u), j, n_v).reshape(n_u*(n_v-2))
        s3 = f(i, torch.remainder(j+1, n_v), n_v).reshape(n_u*(n_v-2))
        s4 = f(torch.remainder(i-1, n_u), j, n_v).reshape(n_u*(n_v-2))
        s5 = f(i, torch.remainder(j-1, n_v), n_v).reshape(n_u*(n_v-2))

        facets1 = torch.stack((s1,s2,s3)).T
        facets2 = torch.stack((s1,s4,s5)).T

        return torch.cat((facets1, facets2))
    

    def draw_surface(self, free_parameters : torch.Tensor, points_per_dim : tuple[int, int]):

        cp = self.free_parameters_to_cp(free_parameters=free_parameters)
        n_u, n_v = points_per_dim

        u, v = (torch.arange(n_u)/n_u)[...,None], (torch.arange(n_v)/(n_v-1))[...,None]

        i, j = torch.arange(0, self.M1)[None,...], torch.arange(-1, self.M2+2)[None, ...]

        phi_u = self.phi_per(self.M1*u-i)[:,None,:,None]
        phi_v = self.phi(self.M2*v-j)[None,:,None,:]

        weights = (phi_u*phi_v).reshape(n_u*n_v, self.M1, self.M2+3)[None,...,None].to(self.device)

        points = (weights*cp).sum(dim=(-2,-3))

        facets = self.create_facets(n_u,n_v)

        values = (torch.cos(2*torch.pi*u)+torch.cos(torch.pi*v).squeeze()[None,...]).reshape(-1)

        return points, facets, values

    
if __name__ == "__main__":

    M = 3
    M1 = 2*M
    M2 = M
    sampler = SnakeSmoothSampler(P = 51, M1 = M1, M2 = M2)

    #cp = torch.rand((1,M1*(M2-1)+6, 3))
    #sampling = sampler.sample_snakes(cp)

    sphere = sampler.sample_sphere()
    #print(sampling.shape)



import hydra
from omegaconf import DictConfig, OmegaConf
import torch
from torch.utils.data import DataLoader
import wandb
import numpy as np
import random
from os import environ
from os.path import join

from nagini3D.data.training_set import TrainingSet, custom_collate
from nagini3D.data.th_optim_set import OptimSet
from nagini3D.models.model import Nagini3D

# reproducibility
def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


@hydra.main(version_base=None, config_path="configs", config_name="train")
def start_training(cfg : DictConfig):
    train(cfg)

def train(cfg : DictConfig):
    device = "cuda" if torch.cuda.is_available() else "cpu"

    # reproducibility
    fixed_seed = cfg["settings"]["seed"]
    g = torch.Generator()
    if type(fixed_seed) == int:
        print(f"Selected seed : {fixed_seed}")
        environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8" 
        torch.manual_seed(fixed_seed)
        random.seed(fixed_seed)
        np.random.seed(fixed_seed)
        torch.use_deterministic_algorithms(True)
        g.manual_seed(fixed_seed)
        

    P = cfg["settings"]["nb_points_on_surface"] # number of points to sample for each snake

    min_cell_ratio = cfg["data"].get("min_cell_ratio")
    if min_cell_ratio is None:
        min_cell_ratio = 0

    patch_size = [int(x) for x in cfg["data"]["patch_size"]]

    anisotropy = [float(x) for x in cfg["data"]["anisotropy"].split(",")]


    train_set = TrainingSet(nb_points=P, patch_size=patch_size, dataset_dir=cfg["data"]["train"],
                            data_aug=cfg["data"]["data_aug"], cell_ratio_th=min_cell_ratio, anisotropy_ratio=anisotropy)
    val_set = TrainingSet(nb_points=P, r_mean=train_set.r_mean,  patch_size=patch_size,\
                            data_aug=cfg["data"]["data_aug"], dataset_dir=cfg["data"]["val"],
                            cell_ratio_th=min_cell_ratio, anisotropy_ratio=anisotropy)


    print(f"Mean radius of cells : {round(train_set.r_mean,3)}")

    train_loader = DataLoader(train_set, collate_fn=custom_collate, batch_size=cfg["settings"]["batch_size"],\
                              worker_init_fn=seed_worker, generator=g)
    val_loader = DataLoader(val_set, collate_fn=custom_collate, batch_size=cfg["settings"]["batch_size"],\
                            worker_init_fn=seed_worker, generator=g)

    nb_epochs = cfg["settings"]["nb_epochs"]

    use_wandb = cfg["settings"]["use_wandb"]


    M1 = cfg["settings"]["M1"]
    M2 = cfg["settings"]["M2"]

    model = Nagini3D(unet_cfg=cfg["model"], P=P, M1=M1, M2=M2,\
                         device=device, save_path=cfg["save"]["path"])
    
    model.init_optimizer(optimizer_cfg=cfg["optimizer"])
    #model.init_scheduler(scheduler_cfg=cfg["scheduler"])
    
    model.init_loss(loss_lambdas=cfg["loss"]["lambdas"], snake_loss=cfg["loss"]["snake_loss"],
                    reg_part=cfg["loss"]["reg_part"], epoch_reg_max=cfg["loss"]["epoch_reg_max"])
    model.set_name_and_save_dir(cfg["settings"]["experiment_name"])

    cfg["settings"] = {**cfg["settings"], "r_mean" : train_set.r_mean}

    model.save_config(cfg_dict=cfg)

    best_val_score = float("inf")
    epoch_best_val = -1

    if use_wandb:
        run = wandb.init(project="SplineDist3D_dense", name=model.get_run_name(), config=dict(cfg))

    for epoch_idx in range(nb_epochs):
        print(f"Epoch {epoch_idx+1} / {nb_epochs}\nTraining ...")

        snake_ratio = model.update_snake_loss(epoch_idx)
        reg_ratio = 1 - snake_ratio

        loss = model.epoch(data_loader = train_loader)

        print(f"\nLoss : {loss['loss']}\nTesting ...")

        validation, cloud_list = model.val(data_loader = val_loader, nb_cells_to_plot=4)
        
        print(f"\nAccuracy on validation set : {validation['loss']}")



        if use_wandb:
            wandb.log({
                "train/snakes" : loss["snakes"], "train/spots" : loss["spots"], "train/loss" : loss["loss"],\
                    "train/reg" : loss["regularization"], "train/variance" : loss["fm_var"],\
                        "test/snakes" : validation["snakes"], "test/spots" : validation["spots"], "test/loss" : validation["loss"],\
                            "test/reg" : validation["regularization"], "test/variance" : validation["fm_var"],\
                                "lr" : model.optimizer.param_groups[0]['lr'], "reg_part" : reg_ratio,\
                "points_clouds" : [
                    wandb.Object3D(point_cloud) for point_cloud in cloud_list
                ]
            })


        if validation["loss"] < best_val_score:
            model.save_model(f"best")
            best_val_score = validation["loss"]
            epoch_best_val = epoch_idx

        
    cfg["save"] = {**cfg["save"], "best_epoch" : epoch_best_val}
    model.save_config(cfg_dict=cfg)
    model.save_model("final")

    wandb.finish()

    model.load_weights(join(model.save_dir, "best.pkl"))
    optim_set = OptimSet(cfg["data"]["val"])
    opti_th = model.optimize_thresholds(optim_set, r_mean=train_set.r_mean, nb_tiles=(1,1,1))
    model.save_th(th_dict=opti_th)

    return model, train_set.r_mean


if __name__ == "__main__":
    start_training()


import torch.nn as nn
from math import log, exp

from .sliced_wasserstein_loss import SlicedWassersteinLoss

def sigmoid_alpha(T_max, reg_init, eps = 1e-4):
    gamma = (2/T_max)*log((1-eps)/eps)

    def f_alpha(x):
        a = exp(-gamma*(x-T_max/2))
        return 1 - reg_init*(a/(a+1))
    
    return f_alpha

def threshold_alpha(T_switch):
    def f_alpha(x):
        return 0 if x<T_switch else 1
    return f_alpha

def exponential_alpha(T_max, reg_init, eps = 1e-3):
    T = -T_max/log(eps)

    def f_alpha(x):
        return 1 - reg_init*exp(-x/T)
    
    return f_alpha


class ShapeReg(nn.Module):
    def __init__(self, shape, device, *args, **kwargs) -> None:
        super().__init__()
        self.shape = shape.unsqueeze(0).to(device)

    def forward(self,x):
        return (x-self.shape).norm(dim=-1).mean()

class RegularizedSnakeLoss(nn.Module):

    def __init__(self, reg_cps, device, f_alpha = "sigmoid", reg_part = 1.0,
                 epoch_reg_max = 200, eps = 1e-3, *args, **kwargs) -> None:
        super().__init__()

        self.shape_reg = ShapeReg(shape = reg_cps, device=device)

        self.device = device

        self.eps = eps

        self.snake_loss = SlicedWassersteinLoss(nb_angles=100, W_order=1, p_order=1, device=self.device)
        
        self.f_alpha = {"sigmoid" : sigmoid_alpha(epoch_reg_max, reg_part),\
                        "exponential" : exponential_alpha(epoch_reg_max, reg_part)}[f_alpha]
        
        self.f_epoch = 0
        self.f_epoch = self.update_reg_factor(0)
        
    def update_reg_factor(self, epoch):
        # stop computing the new value of the multiplication factor if it is close enough to 1
        if self.f_epoch == 1:
            return 1
        self.f_epoch = self.f_alpha(epoch)
        if self.f_epoch > 1 - self.eps:
            self.f_epoch = 1
        return self.f_epoch

    
    def forward(self, sampling, target, proba, cps):

        f = self.f_epoch
        snake = self.snake_loss(sampling, target, proba)
        if f > 1-self.eps:
            return snake
        reg = self.shape_reg(cps)
        return f*snake + (1-f)*reg


if __name__=="__main__":
    T_max = 50
    shift = 0.8
    f = sigmoid_alpha(T_max, shift)
    print(f"f(0) : {f(0)}, f(max) : {f(T_max)}")

    g = exponential_alpha(T_max, shift)
    print(f"g(0) : {g(0)}, g(max) : {g(T_max)}")


from .unet_clean import UNetClean
import torch.nn as nn
from .internals.layers import cInstanceNorm2d, cBlockNorm3d, cBlockNorm3dSmooth, cBlockNorm3dSmoothV2

class DoubleConv(nn.Module):
    """
    [ Conv3d => BatchNorm (optional) => ReLU ] x 2
    """

    def __init__(self, in_ch: int, out_ch: int, train_bn: bool, inner_normalisation: str, padding_mode:str, **kwargs):
        super().__init__()
        inm = {'InstanceNorm':nn.InstanceNorm3d, 'BatchNorm':nn.BatchNorm3d,
                'None' : nn.Identity, 'cInstanceNorm2d' : cInstanceNorm2d,
                'cBlockNorm3d' :cBlockNorm3d, 'cBlockNorm3dSmooth' : cBlockNorm3dSmooth,
                'cBlockNorm3dSmoothV2': cBlockNorm3dSmoothV2}[inner_normalisation]
        self.net = nn.Sequential(
            nn.Conv3d(in_ch, out_ch, kernel_size=3, padding=1, padding_mode=padding_mode),
            inm(out_ch, track_running_stats=train_bn),
            nn.ReLU(inplace=True),
            nn.Conv3d(out_ch, out_ch, kernel_size=3, padding=1, padding_mode=padding_mode),
            inm(out_ch, track_running_stats=train_bn),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.net(x)
    


class DoubleConvRes(nn.Module):
    """
    [ Conv3d => BatchNorm (optional) => ReLU ] x 2
    """

    def __init__(self, in_ch: int, out_ch: int, train_bn: bool, inner_normalisation: str, padding_mode:str, **kwargs):
        super().__init__()
        inm = {'InstanceNorm':nn.InstanceNorm3d, 'BatchNorm':nn.BatchNorm3d,
                'None' : nn.Identity, 'cInstanceNorm2d' : cInstanceNorm2d,
                'cBlockNorm3d' :cBlockNorm3d, 'cBlockNorm3dSmooth' : cBlockNorm3dSmooth,
                'cBlockNorm3dSmoothV2': cBlockNorm3dSmoothV2}[inner_normalisation]
        self.net = nn.Sequential(
            nn.Conv3d(in_ch, out_ch, kernel_size=3, padding=1, padding_mode=padding_mode),
            inm(out_ch, track_running_stats=train_bn),
            nn.ReLU(inplace=True),
            nn.Conv3d(out_ch, out_ch, kernel_size=3, padding=1, padding_mode=padding_mode),
            inm(out_ch, track_running_stats=train_bn),
            nn.ReLU(inplace=True)
        )

        self.residual = nn.Conv3d(in_ch, out_ch, kernel_size=1, padding=0, padding_mode=padding_mode)

    def forward(self, x):
        return self.net(x) + self.residual(x)


class Down(nn.Module):
    """
    Downsampling (by Conv)
    """

    def __init__(self, in_ch: int, out_ch: int, scale_factor: int, padding_mode:str, **kwargs):
        super().__init__()
        self.in_ch = in_ch
        assert out_ch == in_ch, f'Expected out_ch = {in_ch} in Up got out_ch = {out_ch}'
        self.downsample = nn.Conv3d(in_ch, out_ch,
                                    kernel_size=scale_factor,
                                    stride=scale_factor,
                                    padding_mode=padding_mode)

    def forward(self, x):
        """Downsample input by scale factor

        Parameters
        ----------
        x : input tensor ( b, c, t, w, h)

        Returns
        -------
        x : output tensor (b, c, t, w // sf, h // sf)

        """
        assert x.shape[1] ==  self.in_ch, 'Incorrect number of channels'
        return self.downsample(x)

class Up(nn.Module):
    """
    Upsampling (by either bilinear interpolation or transpose convolutions)
    """

    def __init__(self, in_ch: int, out_ch: int, scale_factor: int, padding_mode:str,  bilinear : bool, **kwargs):
        super().__init__()
        assert out_ch == in_ch//scale_factor, f'Expected out_ch = {in_ch//scale_factor} in Up got out_ch = {out_ch}'
        self.upsample = None
        if padding_mode != 'zeros':
             print(f"Could not apply '{padding_mode}' in upsampling used padding_mode='zeros'")
        assert not bilinear, 'Bilinear upsampling not available for 3D Unet'
        # ValueError: Only "zeros" padding mode is supported for ConvTranspose3d
        self.upsample = nn.ConvTranspose3d(in_ch, out_ch,
                                           kernel_size=scale_factor,
                                           stride=scale_factor,
                                           padding_mode='zeros')

    def forward(self, x1):
        return self.upsample(x1)


class Unet3D_3D(UNetClean) :
    lalevels = { 'down' : Down,
                 'convl' : DoubleConv,
                 'transit' : nn.Identity,
                 'convr' : DoubleConv,
                 'up' : Up,
                 'final_layer' : nn.Conv3d}

    def __init__(self, **kwargs) :
        super().__init__(**kwargs)



import torch
import numpy as np
from datetime import datetime
from os.path import join, isdir
from os import mkdir
from omegaconf import OmegaConf
from csbdeep.utils import normalize
from scipy.optimize import minimize_scalar
from tqdm import tqdm
from skimage.segmentation import inverse_gaussian_gradient
from skimage.filters import threshold_otsu, gaussian
from scipy.ndimage import sobel

from .unet.u3D_3D import Unet3D_3D
from .loss.snake_regularized import RegularizedSnakeLoss
from .tools.snake_sampler import SnakeSmoothSampler
from .tools.display_val import PointCloudDisplay
from .tools.coordinates import spherical_to_xyz, init_sphere
from .tools.inference_tools import find_centers, inside_contour, inside_contour_slow, get_mini_maxi
from .tools.cost_matrix import compute_jaccard
from .tools.image_tools import trilinear


class Nagini3D():
    def __init__(self, unet_cfg : dict, P : int = 101, M1 : int  = 4, M2 : int = 2,\
                 device : str = "cpu", save_path : str = ".") -> None:
        self.P = P
        self.M1 = M1
        self.M2 = M2
        self.nb_free_parameters = M1*(M2-1) + 6

        self.device = device

        self.sampler = SnakeSmoothSampler(P,M1,M2,device)
        self.points_displayer = PointCloudDisplay(M1=M1, M2=M2)

        self.model = Unet3D_3D(num_classes = 3*self.nb_free_parameters + 1, **unet_cfg)
        self.model = self.model.to(device)

        self.save_path = save_path

        
    def init_optimizer(self, optimizer_cfg: dict):
        self.optimizer = {"adam" : torch.optim.AdamW, "sgd" : torch.optim.SGD}[optimizer_cfg["name"]](self.model.parameters(), **optimizer_cfg["parameters"])

    def init_loss(self, loss_lambdas, reg_part=0.8, epoch_reg_max = 100):
        self.lambdas = loss_lambdas
        reg_cps = init_sphere(self.M1, self.M2)

        self.criterion = {
            "spots" : torch.nn.BCEWithLogitsLoss(),
            "snakes" : RegularizedSnakeLoss(reg_cps, self.device, f_alpha="sigmoid",
                                            reg_part=reg_part, epoch_reg_max=epoch_reg_max)
        }

    def update_snake_loss(self, epoch):
         return self.criterion["snakes"].update_reg_factor(epoch)

    def set_name_and_save_dir(self, exp_name):
        self.exp_name = exp_name
        now = datetime.now()
        self.run_name = f"{exp_name}_{str(now).replace(' ','_').split('.')[0]}"

        self.save_dir = join(self.save_path, self.run_name)

    def get_run_name(self):
        return self.run_name
    
    def get_save_dir(self):
        return self.save_dir
    
    def save_model(self, model_name : str):
        if not isdir(self.save_dir):
            mkdir(self.save_dir)
        torch.save(self.model.state_dict(), join(self.save_dir, f"{model_name}.pkl"))

    def save_config(self, cfg_dict : dict):
        if not isdir(self.save_dir):
            mkdir(self.save_dir)
        OmegaConf.save(config=cfg_dict, f=join(self.save_dir, "config.yaml"))

    def save_th(self, th_dict : dict):
        if not isdir(self.save_dir):
            mkdir(self.save_dir)
        OmegaConf.save(config=th_dict, f=join(self.save_dir, "thresholds.yaml"))


    def load_weights(self, weights_file):
        """
        Load in the model the weights stored during a training

        Input:
        -weights_file: path to the weights file
        """
        self.model.load_state_dict(torch.load(weights_file, map_location=self.device, weights_only=True))
    
    def normalize_img(self, img):
        return torch.tensor(normalize(img, pmin=1, pmax=99.8, axis=(0,1,2), clip=True), dtype=torch.float32, device=self.device)
    

    def epoch(self, data_loader : torch.utils.data.DataLoader):
        """
        Run a training epoch and return a dictionnary storing the different losses.
        """
        
        infos = {"spots" : 0, "snakes" : 0, "regularization" : 0, "loss" : 0}

        nb_batches = len(data_loader) # Nb of batches

        for idx_batch, batch in enumerate(data_loader):
            print(f"Progression {round(idx_batch/nb_batches*100)}%\r", end="")

            self.optimizer.zero_grad()

            imgs = batch["images"].to(device=self.device)
            proba = batch["proba"].to(device=self.device)
            voxels_proba = batch["voxels_proba"].to(device=self.device)
            cell_voxels = batch["cell_voxels"]

            samplings = batch["voxels_samplings"].to(device=self.device)

            with torch.autocast(device_type="cuda"):

                features, _ = self.model(imgs)

                pred_spots_maps = features[:,0]
                free_parameters = features[:,1:]

                cartesian_list = list()

                for idx_sample, sample in enumerate(free_parameters):
                    x_cell, y_cell, z_cell = cell_voxels[idx_sample]
                    free_param_at_centers = sample[...,x_cell,y_cell, z_cell]

                    x,y,z = spherical_to_xyz(free_param_at_centers)
                    cp = torch.stack((x,y,z)).permute(2,1,0)
                    
                    cartesian_list.append(cp)

                free_parameters_cartesian = torch.cat(cartesian_list)

                pred_samplings = self.sampler.sample_snakes(free_parameters_cartesian)

                bg_mask = (proba==0)[:,None,...]

                spots_loss = self.criterion["spots"](pred_spots_maps, proba)
                loss = spots_loss

                nb_cell_voxel = len(pred_samplings)
                if nb_cell_voxel>0: # do not compute the snake and reg loss if there is no cell voxel in the batch (to avoid NaN)
                    snakes_loss = self.criterion["snakes"](pred_samplings, samplings, voxels_proba, free_parameters_cartesian)
                    snakes_reg = (free_parameters[:,:self.nb_free_parameters]*bg_mask).abs().mean()
                
                    loss = self.lambdas["spots"]*loss + self.lambdas["snakes"]*snakes_loss + self.lambdas["regularization"]*snakes_reg

            loss.backward()
            self.optimizer.step()

            if nb_cell_voxel>0:
                infos["snakes"] += snakes_loss.item()/nb_batches
                infos["regularization"] += snakes_reg.item()/nb_batches
            infos["spots"] += spots_loss.item()/nb_batches
            infos["loss"] += loss.item()/nb_batches
        return infos
    
    def val(self, data_loader : torch.utils.data.DataLoader, nb_cells_to_plot : int = 4):
        """
        Run a validation step and return a dictionnary storing the different losses.

        Input:
        - data_loader: validation data loader,
        - nb_cells_to_plot: if WandB is used, this parameter correspond to the number of 3D point clouds to plot to visualize the accuracy of the predicted surfaces.
        """
        infos = {"spots" : 0, "snakes" : 0, "regularization" : 0, "loss" : 0}
        nb_batches = len(data_loader) # Nb of batches

        with torch.no_grad():
            for idx_batch, batch in enumerate(data_loader):
                print(f"Progression {round(idx_batch/nb_batches*100)}%\r", end="")

                imgs = batch["images"].to(device=self.device)
                proba = batch["proba"].to(device=self.device)
                voxels_proba = batch["voxels_proba"].to(device=self.device)
                cell_voxels = batch["cell_voxels"]

                samplings = batch["voxels_samplings"].to(device=self.device)

                features, _ = self.model(imgs)

                pred_spots_maps = features[:,0]
                free_parameters = features[:,1:]

                cartesian_list = list()

                for idx_sample, sample in enumerate(free_parameters):
                    x_cell, y_cell, z_cell = cell_voxels[idx_sample]
                    free_param_at_centers = sample[...,x_cell,y_cell, z_cell]

                    
                    x,y,z = spherical_to_xyz(free_param_at_centers)
                    cp = torch.stack((x,y,z)).permute(2,1,0)
                    
                    cartesian_list.append(cp)

                free_parameters_cartesian = torch.cat(cartesian_list)

                pred_samplings = self.sampler.sample_snakes(free_parameters_cartesian)

                bg_mask = (proba==0)[:,None,...]

                spots_loss = self.criterion["spots"](pred_spots_maps, proba)
                loss = spots_loss
                
                nb_cell_voxel = len(pred_samplings)
                if nb_cell_voxel>0: # do not compute the snake and reg loss if there is no cell voxel in the batch (to avoid NaN)

                    snakes_loss = self.criterion["snakes"](pred_samplings, samplings, voxels_proba, free_parameters_cartesian)
                    snakes_reg = (free_parameters[:,:self.nb_free_parameters]*bg_mask).abs().mean()
                
                    loss = self.lambdas["spots"]*loss + self.lambdas["snakes"]*snakes_loss + self.lambdas["regularization"]*snakes_reg

                if nb_cell_voxel>0:
                    infos["snakes"] += snakes_loss.item()/nb_batches
                    infos["regularization"] += snakes_reg.item()/nb_batches
                infos["spots"] += spots_loss.item()/nb_batches
                infos["loss"] += loss.item()/nb_batches

            
            idx_pseudo_centers = (voxels_proba == 1)

            cloud_list = self.points_displayer.display_val(nb_cells=nb_cells_to_plot, GT=samplings[idx_pseudo_centers].cpu(),\
                                            pred=pred_samplings[idx_pseudo_centers].cpu(), cp=free_parameters_cartesian[idx_pseudo_centers].cpu())

            return infos, cloud_list
        
    

    def apply_network(self, img, proba_th, GT_centers = None, predict_centers = True, nb_tiles = (1,1,1), overlap_ratio = 0.1):

        """
        Apply the UNet to get the score map and the surface parameters and return the surface parameters at the score map local maxima.
        
        Input:
        - img: np.array(width,height,depth), 3D image to segment
        - proba_th: float, threshold used for object detection (computed automatically on validation set)
        - GT_centers: if provided, will also return the surface parameters at the provided centers
        - predict_centers: bool, if True detect the centers (maxima of the score map)
        - nb_tiles: (int, int,int), number of splits on each dimension to make tiles with the image if it is too big to process in a single time
        - overlap_ratio: float in [0,1], ratio of the tile size to overlap with the previous/next tile 
        
        Output:
        - dict{"proba", "pred"{"centers", "surface_parameters"}, "GT"{"centers", "surface_parameters"}}
        """

        img = self.normalize_img(img)

        Nx, Ny, Nz = img.shape
        nx, ny, nz = nb_tiles

        lx, ly, lz = (Nx//nx) + (Nx%nx > 0), (Ny//ny) + (Ny%ny > 0), (Nz//nz) + (Nz%nz > 0)
        ox, oy, oz = int(lx*overlap_ratio), int(ly*overlap_ratio), int(lz*overlap_ratio)

        centers_list = list()
        parameters_list = list()

        if GT_centers != None:
            GT_param_list = list()
            GT_index_order = list()
            X_GT, Y_GT, Z_GT = GT_centers.T

        proba_pred = torch.zeros_like(img, dtype=float, device=self.device)
        proba_votes = torch.zeros_like(img, dtype=int, device=self.device)

        result = {}

        with torch.no_grad():

            for ix in range(nx):
                X_min, X_max = lx*ix, lx*(ix + 1)
                X_left, X_right = max(0, X_min-ox), min(Nx, X_max+ox)

                for iy in range(ny):
                    Y_min, Y_max = ly*iy, ly*(iy + 1)
                    Y_left, Y_right = max(0, Y_min-oy), min(Ny, Y_max+oy)

                    for iz in range(nz):
                        Z_min, Z_max = lz*iz, lz*(iz + 1)
                        Z_left, Z_right = max(0, Z_min-oz), min(Nz, Z_max+oz)

                        if True :#verbose : 
                            print(f"Infering tile : X : {[X_left,X_right]}, Y : {[Y_left, Y_right]}, Z : {[Z_left, Z_right]}")

                        features, _ = self.model(img[None,None,X_left:X_right,Y_left:Y_right,Z_left:Z_right])

                        pred_proba = torch.sigmoid(features[0,0])
                        free_parameters_spherical = features[0,1:]

                        proba_pred[...,X_left:X_right,Y_left:Y_right,Z_left:Z_right] += pred_proba
                        proba_votes[...,X_left:X_right,Y_left:Y_right,Z_left:Z_right] += 1

                        # Find predicted centers and build the corresponding surfaces
                        if predict_centers:
                            centers = find_centers(pred_proba, proba_th)
                            X,Y,Z = centers.T
                            in_patch = (X>=X_min-X_left)*(X<=X_min-X_left+lx)*\
                                (Y>=Y_min-Y_left)*(Y<=Y_min-Y_left+ly)*\
                                    (Z>=Z_min-Z_left)*(Z<=Z_min-Z_left+lz)
                            X, Y, Z = X[in_patch], Y[in_patch], Z[in_patch]
                            fp_spherical = free_parameters_spherical[..., X, Y, Z]

                            x, y, z = spherical_to_xyz(fp_spherical)

                            fp_cartesian = torch.stack((x,y,z)).permute(2,1,0)

                            centers_list.append(centers[in_patch] + torch.tensor([[X_left, Y_left, Z_left]]).to(self.device))
                            parameters_list.append(fp_cartesian.to(self.device))

                        # If some GT centers where given, build the surfaces at those positions
                        if GT_centers != None:
                            GT_in_patch = (X_GT>=X_min)*(X_GT<=X_max)*\
                                (Y_GT>=Y_min)*(Y_GT<=Y_max)*\
                                    (Z_GT>=Z_min)*(Z_GT<=Z_max)
                            GT_index_order.append(torch.arange(0,len(GT_centers),step=1)[GT_in_patch]) # maybe require a +1 but should not with the new version
                        
                            X_GT_crt, Y_GT_crt, Z_GT_crt = GT_centers[GT_in_patch].T
                            X_GT_shifted, Y_GT_shifted, Z_GT_shifted = X_GT_crt - X_left, Y_GT_crt - Y_left, Z_GT_crt - Z_left

                            GT_spherical = free_parameters_spherical[..., X_GT_shifted, Y_GT_shifted, Z_GT_shifted]
                            x_GT, y_GT, z_GT = spherical_to_xyz(GT_spherical)
                            GT_cartesian = torch.stack((x_GT, y_GT, z_GT)).permute(2,1,0)

                            GT_param_list.append(GT_cartesian.to(self.device))


        # Processing probability prediction
        proba = (proba_pred/proba_votes)

        result["proba"] = proba

        # Processing surface at predicted centers
        if predict_centers:
            all_centers = torch.cat(centers_list, dim=0)
            all_parameters = torch.cat(parameters_list, dim=0)

            prob_centers = proba[all_centers[:,0],all_centers[:,1],all_centers[:,2]]
            _, idx_sort = torch.sort(prob_centers, descending=True)

            all_centers = all_centers[idx_sort]
            all_parameters = all_parameters[idx_sort]

            result["pred"] = {"centers" : all_centers, "surface_parameters" : all_parameters}

        if GT_centers != None:
            all_GT_param = torch.cat(GT_param_list, dim=0)
            all_GT_idx = torch.cat(GT_index_order, dim=0)
            _, order = torch.sort(all_GT_idx)
            all_GT_param = all_GT_param[order]

            result["GT"] = {"centers" : GT_centers, "surface_parameters" : all_GT_param}

        return result
    
    def create_mask(self, img_shape, centers, samplings, facets, nms_th = None, d_mask = 3):

        """
        Create a mask with the same shape as the input image with labels representing the inside of each predicted surfaces.
        
        Input:
        - img_shape: (int,int,int), shape of the initial input image;
        - centers: torch.tensor(n,3), centers of the predicted surfaces
        - samplings: torch.tensor(n,P,3), points sampled on the predicted surfaces
        - facets: torch.tensor(K,3), facets represented by the indexes of the points which are vertices of the facet,
        - nms_th: float in [0,1], Non-Maximum-Suppression threshold, if two object have an IoU greater only one will be kept
        """

        mask_pred = torch.zeros(img_shape, dtype=torch.int32, device=self.device)
        Nx, Ny, Nz = img_shape
        nb_centers = len(centers)
        crt_idx = 1

        remaining_idx = list()
        removed_idx = list()

        with torch.no_grad():
            for i in range(nb_centers):
                s_i = samplings[i] + centers[i,None,:]
                mini, maxi = get_mini_maxi(s_i)
                contour_i = (s_i - mini[None,:] + d_mask)[facets]
                try :
                    patch_mask = inside_contour(contour_i, (maxi-mini+2*d_mask).tolist())
                except :
                    patch_mask = inside_contour_slow(contour_i, (maxi-mini+2*d_mask).tolist())

                patch = patch_mask[d_mask+max(0,-mini[0]):-(d_mask+max(0,maxi[0]-Nx)),
                                   d_mask+max(0,-mini[1]):-(d_mask+max(0,maxi[1]-Ny)),
                                   d_mask+max(0,-mini[2]):-(d_mask+max(0,maxi[2]-Nz))]
                at_patch = mask_pred[max(0,mini[0]):maxi[0], max(0,mini[1]):maxi[1], max(0,mini[2]):maxi[2]]
                fg = at_patch>0
                bg = at_patch==0
                cell_volume = patch.sum()
                overlap = (patch*fg).sum()
                if overlap/cell_volume<nms_th:
                    mask_pred[max(0,mini[0]):maxi[0], max(0,mini[1]):maxi[1], max(0,mini[2]):maxi[2]] += bg*crt_idx*patch
                    crt_idx += 1
                    remaining_idx.append(i)
                else:
                    removed_idx.append(i)

        remaining_centers = centers[remaining_idx]
        removed_centers = centers[removed_idx]

        samp_to_save = (samplings + centers[:,None,:])[remaining_idx]
        
        return mask_pred, samp_to_save, remaining_centers, removed_centers
    

    def optimize_snake(self, img, centers, surf_param, gamma = 0.01, nb_steps = None, up_sampling_ratio = 2, otsu_th : bool = True):

        """
        Apply the snake refinement step on the image using the previously computed surfaces.

        Input:
        - img: np.array(width,height,depth), 3D image used to apply the refinement process
        - centers: np.array(n,3), centers of all the predicted surfaces
        - surf_param: np.array(n,P,3), surface parameters/control points of all the predicted surfaces
        - gamma: float, step of the gradient optimization of the surface parameters
        - n_steps: int(optional), number of steps applied to optimimze the surface parameters (if not provided, computed using gamma)
        - up_sampling_ratio: int>=1, parameter controling the quantity of additional control points, number of initial points is approximately multiplied by up_sampling_ratio^2
        - otsu_th: bool, weither to apply otsu thresholding to img

        Output:
        - tuple(points, facets, values), parameters of the optimized surfaces, points are the position of sampled points on the surface, 
        """

        new_M1, new_M2 = self.M1*up_sampling_ratio, self.M2*up_sampling_ratio
        new_I = new_M1*(new_M2-1)+6
        if up_sampling_ratio>1:
            up_sampler = SnakeSmoothSampler(P = self.P , M1 = new_M1, M2 = new_M2, device=self.device)
            B = self.sampler.up_sample_sampling(surf_param, new_M1, new_M2)
            Theta_inv = up_sampler.up_sample_transfer_matrix()[None,...].expand(len(B),new_I,new_I)

            up_surf_param = torch.bmm(Theta_inv, B)
        else:
            up_sampler = self.sampler
            up_surf_param = surf_param

        if nb_steps == None: nb_steps = int(5/gamma)

        sigmas = [1]
        nb_sig = len(sigmas)

        norm_img = normalize(img, pmin=1, pmax=99.8, axis=(0,1,2), clip=True)
        if otsu_th:
            blurred = gaussian(image=norm_img, sigma=1)
            th = threshold_otsu(blurred)
            otsu = (blurred>th)*1.
            norm_img = otsu


        for sig in sigmas:
            gauss = inverse_gaussian_gradient(norm_img, alpha=100, sigma=sig)

            grad_X = sobel(gauss,0)
            grad_Y = sobel(gauss,1)
            grad_Z = sobel(gauss,2)

            ds_dfi = up_sampler.fi_weights

            for _ in range(nb_steps//nb_sig):
                samples = (up_sampler.sample_snakes(up_surf_param) + centers[:,None,:]).cpu()

                dX = trilinear(samples, grad_X)
                dY = trilinear(samples, grad_Y)
                dZ = trilinear(samples, grad_Z)

                grad = torch.tensor(np.stack((dX,dY,dZ)).transpose((1,2,0))).to(self.device)

                Dfi = (ds_dfi*grad[...,None,:]).sum(dim=(1))

                up_surf_param = up_surf_param - gamma*Dfi

        generated_surf = up_sampler.draw_surface(up_surf_param, points_per_dim=(20,10))
        return up_surf_param, generated_surf
    
    def inference(self, img, proba_th, r_mean, nb_tiles, overlap_ratio = 0.1, nms_th = 0.4, 
                  optim_snake = True, use_otsu = True, d_mask = 3, anisotropy = [1,1,1]):
        """
        Apply the model to a new image.
        
        Input:
        - img: np.array(width,height,depth), 3D image to segment
        - proba_th: float, threshold used for object detection (computed automatically on validation set)
        - r_mean: float, mean radius of the objects used to train the model (stored in the config file during training)
        - nb_tiles: (int, int,int), number of splits on each dimension to make tiles with the image if it is too big to process in a single time
        - overlap_ratio: float in [0,1], ratio of the tile size to overlap with the previous/next tile 
        - nms_th: float in [0,1], Non-Maximum-Suppression threshold, if two object have an IoU greater only one will be kept
        - optim_snake: bool, weither to process snake refinement after network prediction
        - use_otsu: bool, weither to apply Otsu thresholding on image before refinement (improves results if objects are sparse)
        
        Output:
        - mask: np.array(width,height,depth), 3D mask associating a label to each voxel
        - proba: np.array(width,height,depth), 3D map associating a score to be the center of an object to each voxel
        - parameters: dict{"centers", "parameters"}, a dict containing the centers and parameters of the predicted surfaces
        - surfaces: dict{"points", "facets", "values"}, a dict containing the points, facets and values required to reconstruct the surfaces using napari
        """

        with torch.no_grad():
            result = self.apply_network(img, proba_th=proba_th, nb_tiles=nb_tiles, overlap_ratio=overlap_ratio)
        
        anisotropy_vec = np.array(anisotropy)
        anisotropy_ten = torch.tensor(anisotropy_vec, device=self.device)[None,None,...]

        centers, surf_parameters = result["pred"]["centers"], result["pred"]["surface_parameters"]

        if optim_snake:
            surf_parameters, new_surfaces = self.optimize_snake(img, centers, r_mean*surf_parameters/anisotropy_ten, otsu_th=use_otsu)
            samplings, facets, values = new_surfaces
        else:
            samplings, facets, values = self.sampler.draw_surface(free_parameters=r_mean*surf_parameters/anisotropy_ten, points_per_dim=(20,10))
        
        mask, samplings, _, _ = self.create_mask(img.shape, centers=centers, samplings=samplings,\
                                           nms_th=nms_th, facets=facets, d_mask=d_mask)

        return mask.cpu().numpy(), result["proba"].cpu().numpy(), {"centers" : centers, "parameters": surf_parameters},\
            {"points" : samplings.cpu().numpy(), "facets" : facets.cpu().numpy(), "values" : values.cpu().numpy()}
    

    def predict_on_points(self, img, choosen_points, r_mean, nb_tiles, overlap_ratio = 0.1, d_mask = 3):

        result = self.apply_network(img, proba_th=None, r_mean=r_mean, GT_centers=choosen_points, predict_centers=False,\
                                    nb_tiles=nb_tiles, overlap_ratio=overlap_ratio)
        
        mask, _, _, _ = self.create_mask(img.shape, centers=result["GT"]["centers"], samplings=result["GT"]["samplings"],\
                                           nms_th=None, facets=result["tools"]["facets"], d_mask=d_mask)
        
        return mask
    

        
    def optimize_thresholds(self, optim_set, r_mean, nms_th = [0.3,0.5,0.7], nb_tiles = (1,1,1),
                            iou_bins = [0.0,0.2,0.4,0.6,0.8,1.0], maxiter = 20):
        """
        Compute the optimal probability and NSM thresholds on a given dataset.

        Input:
        - optim_set: instance of OptimSet(Dataset), dataset on which to apply the optimization process
        - r_mean: float, mean radius of the objects used to train the model (stored in the config file during training)
        - nb_tiles: (int, int,int), number of splits on each dimension to make tiles with the image if it is too big to process in a single time
        - nms_th: list[float], list of possible values tested for NMS threshold

        Output:
        - optim_th: dict{"prob", "nms"}, dictionnary storing the optim proba and NMS threshold computed
        """
        
        proba_max = max([(self.apply_network(optim_set[n][0], proba_th=0.5, predict_centers=False,\
                                             nb_tiles=nb_tiles,))["proba"].max() for n in range(len(optim_set))])

        proba_max = proba_max.item()

        bounds = proba_max*0.1, proba_max*0.95

        best_nms, best_prob, best_f = None, None, 0  

        for nms in nms_th:
            
            with tqdm(total=maxiter, desc=f"NMS threshold = {nms}") as progress:

                def fn(proba_th):

                    TP, N_GT, N_pred = 0, 0, 0
                    
                    N = len(optim_set)

                    for n in range(N):
                        img, mask = optim_set[n]
                        pred_mask, _, _ = self.inference(img, proba_th, r_mean=r_mean, nb_tiles=nb_tiles,\
                                                        nms_th=nms, verbose=False)
                        tp, n_GT, n_pred = compute_jaccard(mask, pred_mask.detach().cpu().numpy(), iou_bins)
                        TP += tp
                        N_GT += n_GT
                        N_pred += n_pred

                    jaccard = TP / (N_GT + N_pred - TP)
                    iou_gaps = np.array(iou_bins[1:])-np.array(iou_bins[:-1])

                    mean_jaccard = (iou_gaps*jaccard/N).sum()

                    progress.update()
                    progress.set_postfix_str(f"prob_th : {proba_th} -> jaccard area : {mean_jaccard}")
                    progress.refresh()

                    return -mean_jaccard # should return the opposite as we use "minimize_scalar" but want to maximize the jaccard

                    
                # opt = minimize_scalar(fn, method='golden', bracket=bracket, tol=tol, bounds=bounds, options={'maxiter': maxiter})
                opt = minimize_scalar(fn, method='bounded', bounds=bounds, options={'maxiter': maxiter})

            opt_prob, opt_f = opt.x, -opt.fun

            if opt_f > best_f:
                best_f, best_nms, best_prob = opt_f, nms, opt_prob


        opt_th = {"prob" : best_prob.item(), "nms" : best_nms}
        self.opt_th = opt_th

        return opt_th

        


import argparse
from glob import glob
from os.path import join, splitext, basename
import tifffile
import numpy as np
from time import time
from torch.cuda import is_available


from nagini3D.data.format_dataset.data_reading_tools import (compute_barycenter, compute_radius, mask_to_contour,
                                bound_box, farthest_point_sampling, distance_to_center)


#PROBA_TYPE = "distance"
PROBA_TYPE = "barycentric"

#SAMPLING_TYPE = "erosion"
SAMPLING_TYPE = "dilation"


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", help="str: direrctory where the segmentation masks are stored")
    parser.add_argument("-o", "--output", help="str: directory where to store the formated sampling")
    parser.add_argument("-n", "--nb_sampling", type=int, default=101,
                        help="int: number of points to sample on mask boundaries")
    parser.add_argument("-s", "--sampling", type=str2bool, nargs="?", default=True, const=True, help="bool: weither the sampling step should be done or not")
    parser.add_argument("-p", "--proba", type=str2bool, nargs="?", default=True, const=True, help="bool: weither the probability step should be done or not")
    parser.add_argument("-v", "--verbose", type=str2bool, nargs="?", default=False, const=True, help="bool: weither to print logs of the sampling procedure")
    parser.add_argument("-a", "--anisotropy", default="1,1,1", help="str: If your images are strongly anisotropic, precise the anisotropy ratio between axis (with float greater or equal than 1, separated by commas, ex:1,3,1)")

    args = parser.parse_args()

    input_dir = args.input
    output_dir = args.output
    nb_sampling = args.nb_sampling
    do_sample = args.sampling
    do_proba = args.proba
    verbose = args.verbose
    anisotropy = [float(x) for x in args.anisotropy.split(",")]

    device = "cuda" if is_available() else "cpu"

    mask_files = glob(join(input_dir,"*.tif"))

    for mask_f in mask_files: 
        tic = time()
        filename = basename(mask_f)
        name_no_ext = splitext(filename)[0]
        if verbose: print(f"Processing file : {filename}\n")

        mask = tifffile.imread(mask_f)

        nx,ny,nz = mask.shape
        vx,vy,vz = np.arange(nx)[:,None,None], np.arange(ny)[None,:,None], np.arange(nz)[None, None, :]
        mesh = (vx, vy, vz)

        #mesh = np.meshgrid(vx, vy, vz)

        mask_idx = np.unique(mask)[1:]

        nb_cells = len(mask_idx)

        proba_map = np.zeros_like(mask, dtype=float)
        gaussian_mask = np.zeros_like(mask, dtype=int)

        center_list = list()
        radius_list = list()
        sampling_list = list()
        idx_list = list()

        cells_count = 0

        for i,idx in enumerate(mask_idx):
            crt_mask = (mask == idx)*1
            N = crt_mask.sum()
            if verbose: print(f"\rCell nb {idx}/{nb_cells}", end="")
        
            barycenter = compute_barycenter(crt_mask, mesh)

            crt_contour = mask_to_contour(crt_mask, mode = SAMPLING_TYPE)

            radius = compute_radius(crt_contour, mesh, barycenter)

            bb = bound_box(crt_contour, mesh)

            x_min, x_max, y_min, y_max, z_min, z_max = bb

            small_contour = crt_contour[x_min:x_max+1, y_min:y_max+1, z_min:z_max+1]
            small_mask = crt_mask[x_min:x_max+1, y_min:y_max+1, z_min:z_max+1]

            if do_proba:
                if PROBA_TYPE == "distance":
                    distance_map = distance_to_center(mask=small_mask)
                    proba_map[x_min:x_max+1, y_min:y_max+1, z_min:z_max+1] = np.maximum(distance_map, proba_map[x_min:x_max+1, y_min:y_max+1, z_min:z_max+1])
                

                if PROBA_TYPE=="barycentric":
                    bx, by, bz = barycenter
                    dist = np.sqrt((vx-bx)**2+(vy-by)**2+(vz-bz)**2)
                    masked_dist =  dist*crt_mask
                    M = np.max(masked_dist)
                    masked_dist = (M - masked_dist)*crt_mask
                    den = np.max(masked_dist)
                    if den>0:
                        masked_dist = masked_dist/den
                        proba_map += masked_dist

            if do_sample:
                sampling = farthest_point_sampling(small_contour, nb_sampling, device = device, anisotropy=anisotropy)
                centered_sampling = np.array(sampling) + np.array([x_min, y_min, z_min]) - np.array([barycenter])
                sampling_list.append(centered_sampling)

            center_list.append(barycenter)
            radius_list.append(radius)
            idx_list.append(cells_count)

            cells_count+=1


        tac = time()

        if verbose: print(f"\nTime spent to process file : {round(tac-tic)}s")

        if do_sample:
            npz_path = join(output_dir, name_no_ext+".npz")
            np.savez(npz_path, centers = np.array(center_list), radius = np.array(radius_list),\
                 samplings = np.array(sampling_list))
        
        if do_proba:
            proba_path = join(output_dir, name_no_ext+".tif")
            tifffile.imwrite(proba_path, proba_map)



import torch
import numpy as np
from scipy.optimize import linear_sum_assignment

K = 10
BINS = [k/K for k in range(K+1)]


def compute_cost_matrix(GT_mask, pred_mask):
    f_GT = GT_mask.flatten()
    f_pred = pred_mask.flatten()
    M_GT = f_GT.max()
    M_pred = f_pred.max()

    intersection = np.zeros((M_GT+1,M_pred+1))
    S_GT = np.zeros((M_GT+1))
    S_pred = np.zeros((M_pred+1))

    I = len(f_GT)

    for i in range(I):
        intersection[f_GT[i],f_pred[i]] += 1
        S_GT[f_GT[i]] +=1
        S_pred[f_pred[i]]+= 1


    union = S_GT[:,None] + S_pred[None,:] - intersection
    iou = intersection/union

    iou_no_bg = iou[1:,1:]

    return iou_no_bg


def compute_jaccard(GT_mask, pred_mask, bins=BINS):
    C = compute_cost_matrix(GT_mask, pred_mask)
    N_GT, N_pred = C.shape
    row_ind, col_ind = linear_sum_assignment(cost_matrix=C, maximize=True)
    iou = C[row_ind, col_ind]
    res, _ = np.histogram(iou, bins)
    TP = np.cumsum(res[::-1])[::-1]

    return TP, N_GT, N_pred


def iou_at_centers(GT_mask, pred_mask):

    cells_iou = list()
    idx_cells = torch.unique(GT_mask)[1:]

    for k,idx in enumerate(idx_cells):

        crt_GT = (GT_mask == idx)
        crt_pred = (pred_mask == idx)

        inter = (crt_GT*crt_pred).sum()
        union = crt_GT.sum() + crt_pred.sum() - inter

        IoU = inter/union

        cells_iou.append(IoU.item())

    return np.array(cells_iou)


from torch.utils.data import Dataset
from os.path import join, basename, splitext
from glob import glob
from tifffile import imread
from csbdeep.utils import normalize


class OptimSet(Dataset):
    def __init__(self, dataset_dir) -> None:
        super().__init__()
        self.dataset_dir = dataset_dir

        self.images_dir = join(dataset_dir, "images")
        self.masks_dir = join(dataset_dir, "masks")

        self.imgs_list = glob(join(self.images_dir, "*.tif"))

    def __len__(self):
        return len(self.imgs_list)

    def __getitem__(self, index):
        img_path = self.imgs_list[index]
        img_name = basename(img_path)

        full_img = imread(img_path)
        full_img = normalize(full_img, pmin=1, pmax=99.8, axis=(0,1,2), clip=True)

        full_mask = imread(join(self.masks_dir, img_name))
        
        return full_img, full_mask


from math import cos, sin
import numpy as np
from skimage.filters import gaussian
from random import random

def generate_Rx(theta):
    return np.array([
        [1, 0, 0],
        [0, cos(theta), -sin(theta)],
        [0, sin(theta), cos(theta)]
    ])

def generate_Ry(theta):
    return np.array([
        [cos(theta), 0, sin(theta)],
        [0, 1, 0],
        [-sin(theta), 0, cos(theta)]
    
    ])

def generate_Rz(theta):
    return np.array([
        [cos(theta), -sin(theta), 0],
        [sin(theta), cos(theta), 0],
        [0, 0, 1],
    ])

def generate_rotation_matrix(theta1, theta2, theta3):
    return generate_Rz(theta3)@generate_Ry(theta2)@generate_Rx(theta1)

def rotate_image_along_one_dim(img, dim1, dim2, angle):
    """
        angle should be in {0,1,2,3} corresponding to a real angle of angle*pi/2
    """
    if angle%2==1:
        img = np.swapaxes(img, axis1 = dim1, axis2 = dim2)
    if angle>0 and angle<3:
        img = np.flip(img, axis=dim1)
    if angle>1:
        img = np.flip(img, axis=dim2)
    return img

def rotate_image(img, angles):
    angle1, angle2, angle3 = angles
    img = rotate_image_along_one_dim(img, 1, 2, angle1)
    img = rotate_image_along_one_dim(img, 2, 0, angle2)
    return rotate_image_along_one_dim(img, 0, 1, angle3)

def blur_image(img, sigma_max):
    return gaussian(img, sigma_max*random(), preserve_range=True)


import torch
from torch.nn.functional import conv3d

### Creation of the convolution kernel used for fast-marching computation
conv_kernel = torch.zeros((26, 1, 3, 3, 3))
for p in range(26):
    q = p + 1*(p>12)
    i, tmp = q%3, q//3
    j, k = tmp%3, tmp//3
    conv_kernel[p, 0, i,j,k] = -1
    conv_kernel[p, 0, 1,1,1] = 1


def find_centers(spots_map, threshold):
    comparison_map = conv3d(spots_map.float().unsqueeze(0), conv_kernel.to(spots_map.device), padding=1)
    min_map = comparison_map.min(dim=0)[0]
    #max_map = comparison_map.max(dim=0)[0]
    return ((min_map>=0)*(spots_map>threshold)).nonzero()

    
def inside_contours(contours, img_dim):
    """
        - contours : (Batch, nb_facets, nb_points(3), dim_points(3))
    """

    nx, ny, nz = img_dim 
    vx, vy, vz = torch.arange(0, nx), torch.arange(0, ny), torch.arange(0, nz)
    mx, my, mz = torch.meshgrid(vx,vy,vz, indexing="ij")
    voxels = torch.stack((mx,my,mz)).permute(1,2,3,0).to(contours.device)

    v1, v2, v3 = contours[:,:,0], contours[:,:,1], contours[:,:,2] # shape (batch, nb_facets, 3)
    p1, p2 = v1-v2, v3-v2                                          # same
    n = torch.cross(p1,p2,dim=-1)                                         # same
    num = (v2*n).sum(dim=-1)

    B1 = torch.concat((p1[...,None],p2[...,None],n[...,None]), dim = -1) # attention a bien mettre les vecteurs de la base en colonne
    B2 = torch.linalg.inv(B1)[...,:2,:]

    in_map_list = list()

    for i,contour in enumerate(contours):

        in_cell = torch.zeros(size = img_dim).to(contours.device)

        for j, facet in enumerate(contour):

            tmp = n[None,None,None,i,j,:]*voxels
            t = (num[None,None,None,i,j]-tmp[...,2])/tmp[...,:2].sum(dim=-1)

            cdt1 = (t<=1)

            u = torch.concat((voxels[...,:2]*t[...,None], voxels[...,2,None]), dim = -1)

            mat_psg = B2[i,j]

            proj = (mat_psg[None,None,None,...]*(u - v2[None,None,None,i,j])[...,None,:]).sum(dim=-1)

            cdt2 = proj[...,0]>0
            cdt3 = proj[...,1]>0
            cdt4 = proj.sum(dim=-1)<1

            cross_facet = cdt1*cdt2*cdt3*cdt4  

            #print(cross_facet.shape)

            in_cell += cross_facet          
        
        in_map_list.append(in_cell.remainder(2) == 1)
        
    
    return torch.stack(in_map_list)

def inside_contour_slow(contour, patch_dim):

    nx, ny, nz = patch_dim 
    vx, vy, vz = torch.arange(0, nx), torch.arange(0, ny), torch.arange(0, nz)
    mx, my, mz = torch.meshgrid(vx,vy,vz, indexing="ij")
    voxels = torch.stack((mx,my,mz)).permute(1,2,3,0).to(contour.device)

    v1, v2, v3 = contour[:,0], contour[:,1], contour[:,2] # shape (nb_facets, 3)
    p1, p2 = v1-v2, v3-v2                                          # same
    n = torch.cross(p1,p2, dim=-1)                                         # same
    num = (v2*n).sum(dim=-1)

    B1 = torch.concat((p1[...,None],p2[...,None],n[...,None]), dim = -1) # attention a bien mettre les vecteurs de la base en colonne
    B2 = torch.linalg.inv(B1)[...,:2,:]

    in_cell = torch.zeros(size = patch_dim).to(contour.device)

    for j in range(len(contour)):
        tmp = n[None,None,None,j,:]*voxels
        t = (num[None,None,None,j]-tmp[...,2])/tmp[...,:2].sum(dim=-1)

        cdt1 = (t<=1)

        u = torch.concat((voxels[...,:2]*t[...,None], voxels[...,2,None]), dim = -1)

        mat_psg = B2[j]

        proj = (mat_psg[None,None,None,...]*(u - v2[None,None,None,j])[...,None,:]).sum(dim=-1)

        cdt2 = proj[...,0]>=0
        cdt3 = proj[...,1]>=0
        cdt4 = proj.sum(dim=-1)<=1

        cross_facet = cdt1*cdt2*cdt3*cdt4

        in_cell += cross_facet
    
    return (in_cell.remainder(2) == 1)

def inside_contour(contour, patch_dim):
    # faster version that work on a patch of the size if the contour
    nb_facets, _, _ = contour.shape

    nx, ny, nz = patch_dim 
    vx, vy, vz = torch.arange(0, nx), torch.arange(0, ny), torch.arange(0, nz)
    mx, my, mz = torch.meshgrid(vx,vy,vz, indexing="ij")
    voxels = torch.stack((mx,my,mz)).permute(1,2,3,0)[...,None,:].to(contour.device) # shape (mx,my,mz,nb_facets,3)

    v1, v2, v3 = contour[:,0], contour[:,1], contour[:,2] # shape (nb_facets, 3)
    p1, p2 = v1-v2, v3-v2                                          # same
    n = torch.cross(p1,p2, dim=-1)                                         # same
    num = (v2*n).sum(dim=-1)

    B1 = torch.concat((p1[...,None],p2[...,None],n[...,None]), dim = -1) # attention a bien mettre les vecteurs de la base en colonne
    B2 = torch.linalg.inv(B1)[...,:2,:]

    tmp = n[None,None,None,...]*voxels # a voir si ce vecteur n'est pas trop grand
    t = (num[None,None,None,...]-tmp[...,2])/tmp[...,:2].sum(dim=-1)

    u = torch.concat((voxels[...,:2]*t[...,None], voxels[...,2,None].expand(-1,-1,-1,nb_facets,-1)), dim = -1)

    cdt1 = (t<=1) #*(t>=0)

    proj = (B2[None,None,None,...]*(u - v2[None,None,None,...])[...,None,:]).sum(dim=-1)

    cdt2 = proj[...,0]>0
    cdt3 = proj[...,1]>0
    cdt4 = proj.sum(dim=-1)<1

    return (cdt1*cdt2*cdt3*cdt4).sum(dim =-1).remainder(2) == 1


def get_mini_maxi(points):
    mini = points.min(dim=0)[0].floor().to(int)
    maxi = points.max(dim=0)[0].ceil().to(int)

    return mini, maxi









        








import torch
from math import cos, sin, pi

def spherical_to_xyz(free_parameters):

    nb_free_params = free_parameters.shape[0]//3
    
    r = free_parameters[:nb_free_params]
    theta = free_parameters[nb_free_params:2*nb_free_params]*torch.pi
    delta = free_parameters[2*nb_free_params:3*nb_free_params]*torch.pi/2

    x = r*torch.cos(delta)*torch.cos(theta)
    y = r*torch.cos(delta)*torch.sin(theta)
    z = r*torch.sin(delta)

    return x, y, z


def a(M):
    return 2*(1-cos(2*pi/M))/(cos(pi/M)-cos(3*pi/M))

def init_sphere(M1, M2):
    a1 = a(M1)
    a2 = a(2*M2)

    k = torch.arange(start=0, end=M1)
    l = torch.arange(start=1, end=M2)

    x = a1*torch.cos(2*torch.pi*k/M1)[:,None]*a2*torch.sin(torch.pi*l/M2)[None,:]
    y = a1*torch.sin(2*torch.pi*k/M1)[:,None]*a2*torch.sin(torch.pi*l/M2)[None,:]
    z = a2*torch.cos(torch.pi*l/M2).repeat(M1,1)

    reg_cp = torch.stack((x,y,z)).permute(1,2,0).reshape(-1,3)

    phi_0 = (cos(pi/(2*M2))- cos(pi/M2))/(1-cos(pi/M2))
    phi_1 = sin(pi/(4*M2))**2/(1-cos(pi/M2))

    cp_N = torch.tensor([[0,0,a2*(phi_0+2*phi_1*cos(pi/M2))]])
    cp_S = torch.tensor([[0,0,-a2*(phi_0+2*phi_1*cos(pi/M2))]])

    phi_prime_1 = 2*cos(pi/(4*M2))*sin(pi/(4*M2))/(1-cos(pi/M2))
    d = pi*a2*phi_prime_1*sin(pi/M2)

    T_N1 = torch.tensor([[d,0,0]])
    T_N2 = torch.tensor([[0,d,0]])
    T_S1 = torch.tensor([[-d,0,0]])
    T_S2 = torch.tensor([[0,-d,0]])

    return torch.concatenate((reg_cp,cp_N,cp_S,T_N1,T_N2,T_S1,T_S2))


class ControlPointsDimmer():
    def __init__(self, M1, M2, device = "cpu") -> None:
        self.M1 = M1 
        self.M2 = M2

        self.nb_free_params = M1*(M2-1)+6

        self.device = device

        self.sphere_cp = init_sphere(M1,M2).unsqueeze(0).to(device)


    def get_sphere(self):
        return self.sphere_cp
    
    def dim_control_points(self, features):
        dx = features[:self.nb_free_params]
        dy = features[self.nb_free_params:2*self.nb_free_params]
        dz = features[2*self.nb_free_params:3*self.nb_free_params]
        r = features[-1]

        d = torch.stack((dx,dy,dz)).permute(2,1,0)

        return self.sphere_cp*r[:,None,None] + d